"""
This ``Python`` module defines two global reinforcement learning environments that inherit from the
`GraphEnvironment` class. These environments model graph building games in which the edges (resp.
arcs) are initially fully colored in some predetermined manner, and at each step, any edge (resp.
arc) can be properly recolored with any color.
"""

from typing import Optional

import numpy as np

from ..graphs.graph import Graph
from ..graphs.graph_formats import ColorRepresentation, FlattenedOrdering, GraphFormat
from ..graphs.special_graphs import MonochromaticGraph
from ..graphs.utils import graph_order_to_flattened_length
from .graph_environment import (
    EpisodeStatus,
    GraphEnvironment,
    GraphInvariant,
    GraphInvariantDiff,
)
from .graph_generators import GraphGenerator, create_fixed_graph_generator


class GlobalSetEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game in which
    the edges (resp. arcs) are initially fully colored in some manner, and at each step, any edge
    (resp. arc) can be potentially recolored with any color. Users can configure the graph order,
    the number of proper edge colors, whether the graphs are directed or undirected, whether loops
    are allowed, and the mechanism that controls how the initial fully colored graphs are
    generated, which can be deterministic or nondeterministic.

    The RL tasks in this environment are continuing, and the total number of actions to be
    performed, i.e., the episode length, is configurable.

    Each state is represented by a binary `numpy.ndarray` vector of type `numpy.uint8` and length
    ``(edge_colors - 1) * flattened_length``, where ``edge_colors`` is the configured number of
    proper edge colors, and ``flattened_length`` is the flattened length of the graphs. In the
    state vectors, the first ``flattened_length`` bits indicate which edges (resp. arcs) currently
    have color 1, the next ``flattened_length`` bits indicate which edges (resp. arcs) currently
    have color 2, and this pattern continues up to color ``edge_colors - 1``. The edges (resp.
    arcs) are ordered according to the selected `FlattenedOrdering` (row-major or clockwise).

    Each action is represented by a `numpy.int32` integer between 0 and
    ``edge_colors * flattened_length - 1``. Given an action number ``a``, the edge (resp. arc) to
    be recolored is determined by ``a % flattened_length``, and the color to assign is determined
    by ``a // flattened_length``.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _edge_colors: The number of proper edge colors in the graphs to be constructed, given as
        a positive `int` that is at least 2.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying the edge
        (resp. arc) ordering (row-major or clockwise).
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed.
    :ivar _state_length: A positive `int` equal to ``(_edge_colors - 1) * _flattened_length``,
        i.e., the length of each state vector.
    :ivar _episode_length: A positive `int` specifying the episode length, i.e., the total number
        of actions in each episode.
    :ivar _step_count: Either `None` or a nonnegative `int` counting how many actions have been
        executed in the current batch of episodes. When `_step_count` equals `_episode_length`, the
        episode has reached a final state. This attribute is updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        episode_length: Optional[int] = None,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `GlobalSetEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param episode_length: Either `None`, or a positive `int` specifying the number of actions
            in each episode. If `None`, the episode length defaults to the flattened length of the
            graphs to be constructed. The default value is `None`.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying
            whether the edges (resp. arcs) are ordered row-major or clockwise. The default value is
            `FlattenedOrdering.ROW_MAJOR`.
        :param edge_colors: A positive `int` (not below 2) specifying the number of proper edge
            colors in the graphs to be constructed. The default value is 2.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._edge_colors: int = edge_colors
        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )
        self._state_length: int = (self._edge_colors - 1) * self._flattened_length

        if episode_length is not None:
            self._episode_length: int = episode_length
        # If the ``episode_length`` argument is `None`, then the episode length should match the
        # ``_flattened_length`` attribute.
        else:
            self._episode_length: int = self._flattened_length

        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._state_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        return self._edge_colors * self._flattened_length

    @property
    def action_mask(self) -> Optional[np.ndarray]:  # pragma: no cover
        return None

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._episode_length

    @episode_length.setter
    def episode_length(self, episode_length: int):  # pragma: no cover
        """
        This setter allows the user to potentially reconfigure the episode length between two
        independent batches of episodes. It should not be used while a batch of episodes is
        currently in progress.

        :param episode_length: A positive `int` specifying the new episode length.
        """

        self._episode_length = episode_length

    @property
    def is_continuing(self) -> bool:
        return True

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        # Initialize the state vectors using the generated underlying fully colored graphs.
        self._state_batch = (
            format_representation[:, -self._edge_colors + 1 :, :].reshape(batch_size, -1).copy()
        )
        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        rows = np.arange(self._state_batch.shape[0])

        recolored_edges = action_batch % self._flattened_length
        new_colors = action_batch // self._flattened_length

        # If the graphs have only two proper edge colors, then the transition can easily be done as
        # follows.
        if self._edge_colors == 2:
            self._state_batch[rows, recolored_edges] = new_colors
        # Otherwise, the next trick should be used.
        else:
            temp = self._state_batch.reshape(-1, self._edge_colors - 1, self._flattened_length)
            temp[rows, :, recolored_edges] = 0
            temp[rows, new_colors - 1, recolored_edges] = new_colors != 0

        self._step_count += 1
        if self._step_count >= self._episode_length:
            self._status = EpisodeStatus.TRUNCATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        return Graph.from_flattened(
            flattened=state_batch.reshape(
                -1, self._edge_colors - 1, self._flattened_length
            ).copy(),
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            edge_colors=self._edge_colors,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )


class GlobalFlipEnvironment(GraphEnvironment):
    """
    This class inherits from `GraphEnvironment` and models a graph building game for constructing
    2-edge-colored looped complete graphs. In this environment, all edges (resp. arcs) are
    initially fully colored in some manner, and at each step, any edge (resp. arc) can be
    potentially flipped. Users can configure the graph order, whether the graphs are directed or
    undirected, whether loops are allowed, and the mechanism for generating the initial fully
    colored graphs, which can be deterministic or nondeterministic.

    The RL tasks in this environment are continuing. The total number of actions to be performed,
    i.e., the episode length, is configurable.

    Each state is represented by a binary `numpy.ndarray` vector of type `numpy.uint8` that encodes
    the current graph in one of the two flattened formats with color numbers. The edge ordering
    (row-major or clockwise) is configurable by the user at initialization and remains fixed.

    The environment supports two modes of operation controlled by the ``flip_only`` parameter. If
    ``flip_only`` is `False` (the default), each action is a `numpy.int32` integer between 0 and
    ``2 * flattened_length - 1``. In this case, for an action ``a``, ``a % flattened_length`` gives
    the index of the edge (resp. arc) to potentially flip, and ``a // flattened_length`` is a
    binary value indicating whether the edge should actually be flipped (1) or left unchanged (0).
    If ``flip_only`` is `True`, each action is a `numpy.int32` integer between 0 and
    ``flattened_length - 1``, specifying the index of the edge (resp. arc) that must be flipped.

    :ivar _state_batch: See `GraphEnvironment._state_batch`.
    :ivar _status: See `GraphEnvironment._status`.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _flip_only: A `bool` specifying whether the selected edges (resp. arcs) must be flipped.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying the edge
        (resp. arc) ordering (row-major or clockwise).
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed.
    :ivar _episode_length: A positive `int` specifying the episode length, i.e., the total number
        of actions in each episode.
    :ivar _step_count: Either `None` or a nonnegative `int` counting how many actions have been
        executed in the current batch of episodes. When `_step_count` equals `_episode_length`, the
        episode has reached a final state. This attribute is updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        episode_length: Optional[int] = None,
        flip_only: bool = False,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `GlobalFlipEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param episode_length: Either `None`, or a positive `int` specifying the number of actions
            in each episode. If `None`, the episode length defaults to the flattened length of the
            graphs to be constructed. The default value is `None`.
        :param flip_only: A `bool` specifying whether the selected edges (resp. arcs) must be
            flipped. The default value is `False`.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying
            whether the edges (resp. arcs) are ordered row-major or clockwise. The default value is
            `FlattenedOrdering.ROW_MAJOR`.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._flip_only: bool = flip_only
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )

        if episode_length is not None:
            self._episode_length: int = episode_length
        # If the ``episode_length`` argument is `None`, then the episode length should match the
        # ``_flattened_length`` attribute.
        else:
            self._episode_length: int = self._flattened_length

        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._flattened_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        if self._flip_only:
            return self._flattened_length
        else:
            return 2 * self._flattened_length

    @property
    def action_mask(self) -> Optional[np.ndarray]:  # pragma: no cover
        return None

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._episode_length

    @episode_length.setter
    def episode_length(self, episode_length: int):  # pragma: no cover
        """
        This setter allows the user to potentially reconfigure the episode length between two
        independent batches of episodes. It should not be used while a batch of episodes is
        currently in progress.

        :param episode_length: A positive `int` specifying the new episode length.
        """

        self._episode_length = episode_length

    @property
    def is_continuing(self) -> bool:
        return True

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        # Initialize the state vectors using the generated underlying fully colored graphs.
        self._state_batch = format_representation[:, -1, :].reshape(batch_size, -1).copy()
        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        rows = np.arange(self._state_batch.shape[0])
        recolored_edges = action_batch % self._flattened_length

        # If the ``_flip_only`` attribute is `True`, then all the selected edges (resp. arcs) must
        # be flipped.
        if self._flip_only:
            self._state_batch[rows, recolored_edges] ^= 1
        # Otherwise, each selected edge (resp. arc) is either flipped or not flipped, depending on
        # the value ``a // flattened_length``.
        else:
            to_flip_or_not_to_flip = action_batch // self._flattened_length
            self._state_batch[rows, recolored_edges] ^= to_flip_or_not_to_flip.astype(np.uint8)

        self._step_count += 1
        if self._step_count >= self._episode_length:
            self._status = EpisodeStatus.TRUNCATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        return Graph.from_flattened(
            flattened=state_batch.reshape(-1, 1, self._flattened_length).copy(),
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )



"""
The `rlgt.environments` package defines the `GraphEnvironment` abstract base class, which
formalizes a reinforcement learning environment for graph theory applications. It also provides
seven concrete environment classes that inherit from `GraphEnvironment` and implement specific
graph building dynamics. In addition, the package defines the `EpisodeStatus` enumeration, which
describes the execution status of a batch of episodes, and includes several utility functions for
constructing graph generator functions used to initialize batches of underlying graphs.
"""

from .global_environments import GlobalFlipEnvironment, GlobalSetEnvironment
from .graph_environment import (
    EpisodeStatus,
    GraphEnvironment,
    GraphInvariant,
    GraphInvariantDiff,
)
from .graph_generators import (
    GraphGenerator,
    create_choose_two_graph_generator,
    create_edge_perturbation_graph_generator,
    create_fixed_graph_generator,
    create_random_graph_generator,
)
from .linear_environments import (
    LinearBuildEnvironment,
    LinearFlipEnvironment,
    LinearSetEnvironment,
)
from .local_environments import LocalFlipEnvironment, LocalSetEnvironment


__all__ = [
    "GlobalFlipEnvironment",
    "GlobalSetEnvironment",
    "EpisodeStatus",
    "GraphEnvironment",
    "GraphInvariant",
    "GraphInvariantDiff",
    "GraphGenerator",
    "create_choose_two_graph_generator",
    "create_edge_perturbation_graph_generator",
    "create_fixed_graph_generator",
    "create_random_graph_generator",
    "LinearBuildEnvironment",
    "LinearFlipEnvironment",
    "LinearSetEnvironment",
    "LocalFlipEnvironment",
    "LocalSetEnvironment",
]



"""
This ``Python`` module contains the `Graph` class. This class encapsulates the concept of a
*k*-edge-colored looped complete graph or a batch of *k*-edge-colored looped complete graphs of the
same order.
"""

from __future__ import annotations

from typing import Optional

import numpy as np

from .graph_formats import (
    BitmaskType,
    ColorRepresentation,
    FlattenedOrdering,
    GraphFormat,
)
from .utils import (
    binary_slices_to_color_numbers,
    color_numbers_to_binary_slices,
    flatten_from_adjacency_matrix,
    flattened_length_to_graph_order,
    unflatten_to_adjacency_matrix,
)


class Graph:
    r"""
    This class encapsulates the concept of a *k*-edge-colored looped complete graph or a batch of
    *k*-edge-colored looped complete graphs of the same order, with *k* being at least 2. If a
    single graph is modeled, then it is represented as a quintuple ``(edge_colors, is_directed,
    allow_loops, graph_format, format_representation)``, as explained in the `GraphFormat`
    enumeration. If a batch of graphs is modeled, then the graphs in the batch must be consistent
    in the sense that they all have the same number of proper edge colors and the same order, they
    are either all directed or all undirected, and they either all allow loops or all disallow
    them. Accordingly, a batch of graphs can be analogously represented by a quintuple
    ``(edge_colors, is_directed, allow_loops, graph_format, format_representation)``, where the
    only difference is that all `numpy.ndarray` objects used as format representations have one
    additional leading dimension. More precisely, if ``a`` is the `numpy.ndarray` describing the
    structure of a batch of graphs in a given graph format, then ``a[i]`` is the structural
    representation of the *i*-th graph in the batch in that format.

    The structure of the graph or batch of graphs can be represented in any of the eight graph
    formats defined in the `GraphFormat` enumeration. The structure is initialized in at least one
    of these formats so that the corresponding ``format_representation`` `numpy.ndarray` objects
    are internally stored. Afterwards, the graph or batch of graphs can be accessed in any of the
    eight formats, with all required graph format conversions being performed automatically. Each
    time a conversion is performed, the resulting `numpy.ndarray` corresponding to the target graph
    format is internally stored. The class provides support for all reduced formats, and when
    converting to a bitmask format or to a format with binary slices, the corresponding reduced
    format is always used whenever possible, i.e., whenever the graph or batch of graphs is fully
    colored.

    :ivar __edge_colors: The number of proper edge colors, i.e., *k*, given as a positive `int`
        that is at least 2.
    :ivar __is_directed: A `bool` indicating whether the graph or each graph in the batch is a
        *k*-edge-colored looped complete directed graph or a *k*-edge-colored looped complete
        undirected graph.
    :ivar __allow_loops: A `bool` indicating whether the graph or each graph in the batch is
        allowed to have loops. If loops are not allowed, then all loops are removed from the
        considered looped complete graph or graphs and therefore do not exist.
    :ivar __batch_size: Either `None`, if the object models a single graph, or a positive `int`, if
        the object models a batch of graphs, specifying the batch size, i.e., the number of graphs
        in the batch.
    :ivar __graph_order: The graph order or the common order of all graphs in the batch, given as a
        positive `int`.
    :ivar __bitmask_out: Either a `numpy.ndarray` of type `numpy.uint64` representing the graph
        structure in the bitmask format for the out-neighborhoods, if this format was used during
        initialization or computed afterwards, or `None` otherwise.
    :ivar __bitmask_in: Either a `numpy.ndarray` of type `numpy.uint64` representing the graph
        structure in the bitmask format for the in-neighborhoods, if this format was used during
        initialization or computed afterwards, or `None` otherwise.
    :ivar __adjacency_matrix_colors: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the adjacency matrix format with color numbers, if this format was
        used during initialization or computed afterwards, or `None` otherwise.
    :ivar __adjacency_matrix_binary: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the adjacency matrix format with binary slices, if this format was
        used during initialization or computed afterwards, or `None` otherwise.
    :ivar __flattened_row_major_colors: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the flattened row-major format with color numbers, if this format
        was used during initialization or computed afterwards, or `None` otherwise.
    :ivar __flattened_row_major_binary: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the flattened row-major format with binary slices, if this format
        was used during initialization or computed afterwards, or `None` otherwise.
    :ivar __flattened_clockwise_colors: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the flattened clockwise format with color numbers, if this format
        was used during initialization or computed afterwards, or `None` otherwise.
    :ivar __flattened_clockwise_binary: Either a `numpy.ndarray` of type `numpy.uint8` representing
        the graph structure in the flattened clockwise format with binary slices, if this format
        was used during initialization or computed afterwards, or `None` otherwise.
    """

    def __init__(
        self,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
        bitmask_out: Optional[np.ndarray] = None,
        bitmask_in: Optional[np.ndarray] = None,
        adjacency_matrix_colors: Optional[np.ndarray] = None,
        adjacency_matrix_binary: Optional[np.ndarray] = None,
        flattened_row_major_colors: Optional[np.ndarray] = None,
        flattened_row_major_binary: Optional[np.ndarray] = None,
        flattened_clockwise_colors: Optional[np.ndarray] = None,
        flattened_clockwise_binary: Optional[np.ndarray] = None,
    ):
        """
        This method initializes an instance of the class, which models either a single graph or a
        batch of graphs in at least one of the eight graph formats defined in the `GraphFormat`
        enumeration. In other words, at least one graph format must be selected, and the graph or
        batch of graphs is initialized in each selected format using the corresponding provided
        `numpy.ndarray` objects. If the instance is initialized in more than one graph format, then
        the provided `numpy.ndarray` objects must be mutually consistent, meaning that they must
        all represent the same graph or the same batch of graphs.

        :param edge_colors: A positive `int` that represents the number of proper edge colors,
            i.e., *k*, and that must be at least 2. The default value is 2.
        :param is_directed: A `bool` indicating whether the graph or each graph in the batch is a
            *k*-edge-colored looped complete directed graph or a *k*-edge-colored looped complete
            undirected graph. The default value is `False`, meaning that the graph or batch of
            graphs is undirected by default.
        :param allow_loops: A `bool` indicating whether the graph or each graph in the batch is
            allowed to have loops. If loops are not allowed, then all loops are removed from the
            considered looped complete graph or graphs and therefore do not exist. The default
            value is `False`, meaning that loops are not allowed by default.
        :param bitmask_out: Either a `numpy.ndarray` of type `numpy.uint64` describing the graph
            structure in the bitmask format for the out-neighborhoods, if the graph or batch of
            graphs should be initialized in this format, or `None` otherwise. The default value is
            `None`.
        :param bitmask_in: Either a `numpy.ndarray` of type `numpy.uint64` describing the graph
            structure in the bitmask format for the in-neighborhoods, if the graph or batch of
            graphs should be initialized in this format, or `None` otherwise. The default value is
            `None`.
        :param adjacency_matrix_colors: Either a `numpy.ndarray` of type `numpy.uint8` describing
            the graph structure in the adjacency matrix format with color numbers, if the graph or
            batch of graphs should be initialized in this format, or `None` otherwise. The default
            value is `None`.
        :param adjacency_matrix_binary: Either a `numpy.ndarray` of type `numpy.uint8` describing
            the graph structure in the adjacency matrix format with binary slices, if the graph or
            batch of graphs should be initialized in this format, or `None` otherwise. The default
            value is `None`.
        :param flattened_row_major_colors: Either a `numpy.ndarray` of type `numpy.uint8`
            describing the graph structure in the flattened row-major format with color numbers, if
            the graph or batch of graphs should be initialized in this format, or `None` otherwise.
            The default value is `None`.
        :param flattened_row_major_binary: Either a `numpy.ndarray` of type `numpy.uint8`
            describing the graph structure in the flattened row-major format with binary slices, if
            the graph or batch of graphs should be initialized in this format, or `None` otherwise.
            The default value is `None`.
        :param flattened_clockwise_colors: Either a `numpy.ndarray` of type `numpy.uint8`
            describing the graph structure in the flattened clockwise format with color numbers, if
            the graph or batch of graphs should be initialized in this format, or `None` otherwise.
            The default value is `None`.
        :param flattened_clockwise_binary: Either a `numpy.ndarray` of type `numpy.uint8`
            describing the graph structure in the flattened clockwise format with binary slices, if
            the graph or batch of graphs should be initialized in this format, or `None` otherwise.
            The default value is `None`.
        """

        self.__edge_colors: int = edge_colors
        self.__is_directed: bool = is_directed
        self.__allow_loops: bool = allow_loops

        # Initialize the graph or batch of graphs in all selected graph formats.
        self.__bitmask_out: Optional[np.ndarray] = bitmask_out
        self.__bitmask_in: Optional[np.ndarray] = bitmask_in
        self.__adjacency_matrix_colors: Optional[np.ndarray] = adjacency_matrix_colors
        self.__adjacency_matrix_binary: Optional[np.ndarray] = adjacency_matrix_binary
        self.__flattened_row_major_colors: Optional[np.ndarray] = flattened_row_major_colors
        self.__flattened_row_major_binary: Optional[np.ndarray] = flattened_row_major_binary
        self.__flattened_clockwise_colors: Optional[np.ndarray] = flattened_clockwise_colors
        self.__flattened_clockwise_binary: Optional[np.ndarray] = flattened_clockwise_binary

        # If the graph or each graph in the batch is undirected, then the two bitmask formats
        # coincide, so reuse whichever one was provided.
        if not is_directed:
            if bitmask_out is not None:
                self.__bitmask_in = bitmask_out
            elif bitmask_in is not None:
                self.__bitmask_out = bitmask_in

        # Determine the batch size (possibly `None`) and the graph order using the bitmask format
        # for the out-neighborhoods, if available.
        if bitmask_out is not None:
            self.__batch_size: Optional[int] = (
                None if len(bitmask_out.shape) == 2 else bitmask_out.shape[0]
            )
            self.__graph_order: int = bitmask_out.shape[-1]

        # Otherwise, fall back to the bitmask format for the in-neighborhoods.
        elif bitmask_in is not None:
            self.__batch_size: Optional[int] = (
                None if len(bitmask_in.shape) == 2 else bitmask_in.shape[0]
            )
            self.__graph_order: int = bitmask_in.shape[-1]

        # Otherwise, use the adjacency matrix format with color numbers.
        elif adjacency_matrix_colors is not None:
            self.__batch_size: Optional[int] = (
                None
                if len(adjacency_matrix_colors.shape) == 2
                else adjacency_matrix_colors.shape[0]
            )
            self.__graph_order: int = adjacency_matrix_colors.shape[-1]

        # Otherwise, use the adjacency matrix format with binary slices.
        elif adjacency_matrix_binary is not None:
            self.__batch_size: Optional[int] = (
                None
                if len(adjacency_matrix_binary.shape) == 3
                else adjacency_matrix_binary.shape[0]
            )
            self.__graph_order: int = adjacency_matrix_binary.shape[-1]

        # Otherwise, use the flattened row-major format with color numbers.
        elif flattened_row_major_colors is not None:
            self.__batch_size: Optional[int] = (
                None
                if len(flattened_row_major_colors.shape) == 1
                else flattened_row_major_colors.shape[0]
            )
            self.__graph_order: int = flattened_length_to_graph_order(
                flattened_length=flattened_row_major_colors.shape[-1],
                is_directed=is_directed,
                allow_loops=allow_loops,
            )

        # Otherwise, use the flattened row-major format with binary slices.
        elif flattened_row_major_binary is not None:
            self.__batch_size: Optional[int] = (
                None
                if len(flattened_row_major_binary.shape) == 2
                else flattened_row_major_binary.shape[0]
            )
            self.__graph_order: int = flattened_length_to_graph_order(
                flattened_length=flattened_row_major_binary.shape[-1],
                is_directed=is_directed,
                allow_loops=allow_loops,
            )

        # Otherwise, use the flattened clockwise format with color numbers.
        elif flattened_clockwise_colors is not None:
            self.__batch_size: Optional[int] = (
                None
                if len(flattened_clockwise_colors.shape) == 1
                else flattened_clockwise_colors.shape[0]
            )
            self.__graph_order: int = flattened_length_to_graph_order(
                flattened_length=flattened_clockwise_colors.shape[-1],
                is_directed=is_directed,
                allow_loops=allow_loops,
            )

        # Finally, fall back to the flattened clockwise format with binary slices.
        else:
            self.__batch_size: Optional[int] = (
                None
                if len(flattened_clockwise_binary.shape) == 2
                else flattened_clockwise_binary.shape[0]
            )
            self.__graph_order: int = flattened_length_to_graph_order(
                flattened_length=flattened_clockwise_binary.shape[-1],
                is_directed=is_directed,
                allow_loops=allow_loops,
            )

    @classmethod
    def from_bitmask(
        cls,
        bitmask: np.ndarray,
        bitmask_type: BitmaskType = BitmaskType.OUT_NEIGHBORS,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
    ) -> Graph:
        """
        This class method initializes a `Graph` object in either the bitmask format for the
        out-neighborhoods or the bitmask format for the in-neighborhoods.

        :param bitmask: A `numpy.ndarray` of type `numpy.uint64` that describes the format
            representation of the graph or batch of graphs to be initialized, either in the bitmask
            format for the out-neighborhoods or in the bitmask format for the in-neighborhoods.
        :param bitmask_type: An item of the `BitmaskType` enumeration that determines whether the
            graph or batch of graphs is initialized in the bitmask format for the out-neighborhoods
            or in the bitmask format for the in-neighborhoods. The default value is
            `BitmaskType.OUT_NEIGHBORS`, meaning that the graph or batch of graphs is initialized
            in the bitmask format for the out-neighborhoods by default.
        :param edge_colors: A positive `int` that represents the number of proper edge colors,
            i.e., *k*, and that must be at least 2. The default value is 2.
        :param is_directed: A `bool` indicating whether the graph or each graph in the batch is a
            *k*-edge-colored looped complete directed graph or a *k*-edge-colored looped complete
            undirected graph. The default value is `False`, meaning that the graph or batch of
            graphs is undirected by default.
        :param allow_loops: A `bool` indicating whether the graph or each graph in the batch is
            allowed to have loops. If loops are not allowed, then all loops are removed from the
            considered looped complete graph or graphs and therefore do not exist. The default
            value is `False`, meaning that loops are not allowed by default.

        :return: The initialized `Graph` object.

        :note: This class method supports the reduced bitmask formats, meaning that the graph or
            batch of graphs can also be initialized using either of the reduced bitmask format
            representations.
        """

        if bitmask_type == BitmaskType.OUT_NEIGHBORS:
            return cls(
                edge_colors=edge_colors,
                is_directed=is_directed,
                allow_loops=allow_loops,
                bitmask_out=bitmask,
            )
        else:
            return cls(
                edge_colors=edge_colors,
                is_directed=is_directed,
                allow_loops=allow_loops,
                bitmask_in=bitmask,
            )

    @classmethod
    def from_adjacency_matrix(
        cls,
        adjacency_matrix: np.ndarray,
        color_representation: ColorRepresentation = ColorRepresentation.COLOR_NUMBERS,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
    ) -> Graph:
        """
        This class method initializes a `Graph` object in either the adjacency matrix format with
        color numbers or the adjacency matrix format with binary slices.

        :param adjacency_matrix: A `numpy.ndarray` of type `numpy.uint8` that describes the format
            representation of the graph or batch of graphs to be initialized, either in the
            adjacency matrix format with color numbers or in the adjacency matrix format with
            binary slices.
        :param color_representation: An item of the `ColorRepresentation` enumeration that
            determines whether the graph or batch of graphs is initialized in the adjacency matrix
            format with color numbers or in the adjacency matrix format with binary slices. The
            default value is `ColorRepresentation.COLOR_NUMBERS`, meaning that the graph or batch
            of graphs is initialized in the adjacency matrix format with color numbers by default.
        :param edge_colors: A positive `int` that represents the number of proper edge colors,
            i.e., *k*, and that must be at least 2. The default value is 2.
        :param is_directed: A `bool` indicating whether the graph or each graph in the batch is a
            *k*-edge-colored looped complete directed graph or a *k*-edge-colored looped complete
            undirected graph. The default value is `False`, meaning that the graph or batch of
            graphs is undirected by default.
        :param allow_loops: A `bool` indicating whether the graph or each graph in the batch is
            allowed to have loops. If loops are not allowed, then all loops are removed from the
            considered looped complete graph or graphs and therefore do not exist. The default
            value is `False`, meaning that loops are not allowed by default.

        :return: The initialized `Graph` object.

        :note: This class method supports the reduced adjacency matrix format with binary slices,
            meaning that the reduced variant is also accepted when initializing the graph or batch
            of graphs in this format.
        """

        if color_representation == ColorRepresentation.COLOR_NUMBERS:
            return cls(
                edge_colors=edge_colors,
                is_directed=is_directed,
                allow_loops=allow_loops,
                adjacency_matrix_colors=adjacency_matrix,
            )
        else:
            return cls(
                edge_colors=edge_colors,
                is_directed=is_directed,
                allow_loops=allow_loops,
                adjacency_matrix_binary=adjacency_matrix,
            )

    @classmethod
    def from_flattened(
        cls,
        flattened: np.ndarray,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        color_representation: ColorRepresentation = ColorRepresentation.COLOR_NUMBERS,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
    ) -> Graph:
        """
        This class method initializes a `Graph` object in exactly one of the four possible
        flattened graph formats.

        :param flattened: A `numpy.ndarray` of type `numpy.uint8` that describes the format
            representation of the graph or batch of graphs to be initialized, in one of the four
            possible flattened formats.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration that determines
            whether the graph or batch of graphs is initialized in a flattened row-major format or
            in a flattened clockwise format. The default value is `FlattenedOrdering.ROW_MAJOR`,
            meaning that the graph or batch of graphs is initialized in a flattened row-major
            format by default.
        :param color_representation: An item of the `ColorRepresentation` enumeration that
            determines whether the graph or batch of graphs is initialized in a flattened format
            with color numbers or in a flattened format with binary slices. The default value is
            `ColorRepresentation.COLOR_NUMBERS`, meaning that the graph or batch of graphs is
            initialized in a flattened format with color numbers by default.
        :param edge_colors: A positive `int` that represents the number of proper edge colors,
            i.e., *k*, and that must be at least 2. The default value is 2.
        :param is_directed: A `bool` indicating whether the graph or each graph in the batch is a
            *k*-edge-colored looped complete directed graph or a *k*-edge-colored looped complete
            undirected graph. The default value is `False`, meaning that the graph or batch of
            graphs is undirected by default.
        :param allow_loops: A `bool` indicating whether the graph or each graph in the batch is
            allowed to have loops. If loops are not allowed, then all loops are removed from the
            considered looped complete graph or graphs and therefore do not exist. The default
            value is `False`, meaning that loops are not allowed by default.

        :return: The initialized `Graph` object.

        :note: This class method supports the reduced flattened formats with binary slices, meaning
            that the reduced variants are also accepted when initializing the graph or batch of
            graphs in these formats.
        """

        if flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            if color_representation == ColorRepresentation.COLOR_NUMBERS:
                return cls(
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                    flattened_row_major_colors=flattened,
                )
            else:
                return cls(
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                    flattened_row_major_binary=flattened,
                )
        else:
            if color_representation == ColorRepresentation.COLOR_NUMBERS:
                return cls(
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                    flattened_clockwise_colors=flattened,
                )
            else:
                return cls(
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                    flattened_clockwise_binary=flattened,
                )

    @property
    def edge_colors(self) -> int:
        """
        This property returns the number of proper edge colors in the given graph or batch of
        graphs, i.e., *k*, as a positive `int` that is at least 2.
        """

        return self.__edge_colors

    @property
    def is_directed(self) -> bool:
        """
        This property returns a `bool` indicating whether the given graph or each graph in the
        batch is directed. The value `True` indicates that the graph or graphs are directed.
        """

        return self.__is_directed

    @property
    def allow_loops(self) -> bool:
        """
        This property returns a `bool` indicating whether loops are allowed in the given graph or
        in each graph in the batch. The value `True` indicates that loops are allowed.
        """

        return self.__allow_loops

    @property
    def batch_size(self) -> Optional[int]:
        """
        This property returns `None` if the object represents a single *k*-edge-colored looped
        complete graph. If the object represents a batch of *k*-edge-colored looped complete graphs
        of the same order, then this property returns the batch size, i.e., the number of graphs in
        the batch, as a positive `int`.
        """

        return self.__batch_size

    @property
    def graph_order(self) -> int:
        """
        This property returns the order of the given graph or the common order of all graphs in the
        batch, as a positive `int`.
        """

        return self.__graph_order

    def __convert_graph_format(
        self, input_format: GraphFormat, output_format: GraphFormat
    ) -> None:
        """
        This private method performs a direct conversion from the format representation of the
        given graph or batch of graphs in a selected input format to the format representation of
        that graph or batch of graphs in a selected output format. The following direct conversions
        are supported:

        1. bitmask format for the out-neighborhoods ↔ adjacency matrix format with binary slices;
        2. bitmask format for the in-neighborhoods ↔ adjacency matrix format with binary slices;
        3. adjacency matrix format with binary slices ↔ flattened row-major format with binary
           slices;
        4. adjacency matrix format with binary slices ↔ flattened clockwise format with binary
           slices;
        5. adjacency matrix format with color numbers ↔ flattened row-major format with color
           numbers;
        6. adjacency matrix format with color numbers ↔ flattened clockwise format with color
           numbers;
        7. adjacency matrix format with binary slices ↔ adjacency matrix format with color numbers;
        8. flattened row-major format with binary slices ↔ flattened row-major format with color
           numbers; and
        9. flattened clockwise format with binary slices ↔ flattened clockwise format with color
           numbers.

        A `NotImplementedError` is raised if any other direct conversion is attempted.

        :param input_format: The selected input format, given as an item of the `GraphFormat`
            enumeration.
        :param output_format: The selected output format, given as an item of the `GraphFormat`
            enumeration.
        """

        # Bitmask format for the out-neighborhoods → adjacency matrix format with binary slices.
        if input_format == GraphFormat.BITMASK_OUT:
            if output_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
                masks = 1 << np.arange(self.__graph_order, dtype=np.uint64)
                temp = np.expand_dims(self.__bitmask_out, axis=-1) & masks
                self.__adjacency_matrix_binary = (temp != 0).astype(np.uint8)
            else:
                raise NotImplementedError

        # Bitmask format for the in-neighborhoods → adjacency matrix format with binary slices.
        elif input_format == GraphFormat.BITMASK_IN:
            if output_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
                masks = 1 << np.arange(self.__graph_order, dtype=np.uint64)
                temp = np.expand_dims(self.__bitmask_in, axis=-1) & masks
                temp = (temp != 0).astype(np.uint8)
                self.__adjacency_matrix_binary = np.swapaxes(temp, -1, -2)
            else:
                raise NotImplementedError

        # Adjacency matrix format with binary slices → bitmask formats, adjacency matrix format
        # with color numbers, or flattened formats with binary slices.
        elif input_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
            if output_format == GraphFormat.BITMASK_OUT:
                masks = 1 << np.arange(self.__graph_order, dtype=np.uint64)
                self.__bitmask_out = self.__adjacency_matrix_binary @ masks
            elif output_format == GraphFormat.BITMASK_IN:
                masks = 1 << np.arange(self.__graph_order, dtype=np.uint64)
                self.__bitmask_in = np.tensordot(
                    masks, self.__adjacency_matrix_binary, axes=(0, -2)
                )
            elif output_format == GraphFormat.ADJACENCY_MATRIX_COLORS:
                self.__adjacency_matrix_colors = binary_slices_to_color_numbers(
                    input_representation=self.__adjacency_matrix_binary,
                    is_flattened_format=False,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            elif output_format == GraphFormat.FLATTENED_ROW_MAJOR_BINARY:
                self.__flattened_row_major_binary = flatten_from_adjacency_matrix(
                    adjacency_matrix=self.__adjacency_matrix_binary,
                    flattened_ordering=FlattenedOrdering.ROW_MAJOR,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                )
            elif output_format == GraphFormat.FLATTENED_CLOCKWISE_BINARY:
                self.__flattened_clockwise_binary = flatten_from_adjacency_matrix(
                    adjacency_matrix=self.__adjacency_matrix_binary,
                    flattened_ordering=FlattenedOrdering.CLOCKWISE,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

        # Adjacency matrix format with color numbers → adjacency matrix format with binary slices
        # or flattened formats with color numbers.
        elif input_format == GraphFormat.ADJACENCY_MATRIX_COLORS:
            if output_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
                self.__adjacency_matrix_binary = color_numbers_to_binary_slices(
                    input_representation=self.__adjacency_matrix_colors,
                    is_flattened_format=False,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            elif output_format == GraphFormat.FLATTENED_ROW_MAJOR_COLORS:
                self.__flattened_row_major_colors = flatten_from_adjacency_matrix(
                    adjacency_matrix=self.__adjacency_matrix_colors,
                    flattened_ordering=FlattenedOrdering.ROW_MAJOR,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                )
            elif output_format == GraphFormat.FLATTENED_CLOCKWISE_COLORS:
                self.__flattened_clockwise_colors = flatten_from_adjacency_matrix(
                    adjacency_matrix=self.__adjacency_matrix_colors,
                    flattened_ordering=FlattenedOrdering.CLOCKWISE,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

        # Flattened row-major format with binary slices → adjacency matrix format with binary
        # slices or flattened row-major format with color numbers.
        elif input_format == GraphFormat.FLATTENED_ROW_MAJOR_BINARY:
            if output_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
                self.__adjacency_matrix_binary = unflatten_to_adjacency_matrix(
                    flattened=self.__flattened_row_major_binary,
                    flattened_ordering=FlattenedOrdering.ROW_MAJOR,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                    graph_order=self.__graph_order,
                )
            elif output_format == GraphFormat.FLATTENED_ROW_MAJOR_COLORS:
                self.__flattened_row_major_colors = binary_slices_to_color_numbers(
                    input_representation=self.__flattened_row_major_binary,
                    is_flattened_format=True,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

        # Flattened clockwise format with binary slices → adjacency matrix format with binary
        # slices or flattened clockwise format with color numbers.
        elif input_format == GraphFormat.FLATTENED_CLOCKWISE_BINARY:
            if output_format == GraphFormat.ADJACENCY_MATRIX_BINARY:
                self.__adjacency_matrix_binary = unflatten_to_adjacency_matrix(
                    flattened=self.__flattened_clockwise_binary,
                    flattened_ordering=FlattenedOrdering.CLOCKWISE,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                    graph_order=self.__graph_order,
                )
            elif output_format == GraphFormat.FLATTENED_CLOCKWISE_COLORS:
                self.__flattened_clockwise_colors = binary_slices_to_color_numbers(
                    input_representation=self.__flattened_clockwise_binary,
                    is_flattened_format=True,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

        # Flattened row-major format with color numbers → adjacency matrix format with color
        # numbers or flattened row-major format with binary slices.
        elif input_format == GraphFormat.FLATTENED_ROW_MAJOR_COLORS:
            if output_format == GraphFormat.ADJACENCY_MATRIX_COLORS:
                self.__adjacency_matrix_colors = unflatten_to_adjacency_matrix(
                    flattened=self.__flattened_row_major_colors,
                    flattened_ordering=FlattenedOrdering.ROW_MAJOR,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                    graph_order=self.__graph_order,
                )
            elif output_format == GraphFormat.FLATTENED_ROW_MAJOR_BINARY:
                self.__flattened_row_major_binary = color_numbers_to_binary_slices(
                    input_representation=self.__flattened_row_major_colors,
                    is_flattened_format=True,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

        # Flattened clockwise format with color numbers → adjacency matrix format with color
        # numbers or flattened clockwise format with binary slices.
        elif input_format == GraphFormat.FLATTENED_CLOCKWISE_COLORS:
            if output_format == GraphFormat.ADJACENCY_MATRIX_COLORS:
                self.__adjacency_matrix_colors = unflatten_to_adjacency_matrix(
                    flattened=self.__flattened_clockwise_colors,
                    flattened_ordering=FlattenedOrdering.CLOCKWISE,
                    is_directed=self.__is_directed,
                    allow_loops=self.__allow_loops,
                    graph_order=self.__graph_order,
                )
            elif output_format == GraphFormat.FLATTENED_CLOCKWISE_BINARY:
                self.__flattened_clockwise_binary = color_numbers_to_binary_slices(
                    input_representation=self.__flattened_clockwise_colors,
                    is_flattened_format=True,
                    edge_colors=self.__edge_colors,
                    allow_loops=self.__allow_loops,
                )
            else:
                raise NotImplementedError

    @property
    def bitmask_out(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint64` representing the given
        graph or batch of graphs in the bitmask format for the out-neighborhoods.
        """

        # Return the stored array if available.
        if self.__bitmask_out is not None:
            return self.__bitmask_out

        # Compute from the adjacency matrix format with binary slices.
        self.adjacency_matrix_binary
        self.__convert_graph_format(
            input_format=GraphFormat.ADJACENCY_MATRIX_BINARY, output_format=GraphFormat.BITMASK_OUT
        )

        return self.__bitmask_out

    @property
    def bitmask_in(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint64` representing the given
        graph or batch of graphs in the bitmask format for the the in-neighborhoods.
        """

        # Return the stored array if available.
        if self.__bitmask_in is not None:
            return self.__bitmask_in

        # Compute from the adjacency matrix format with binary slices.
        self.adjacency_matrix_binary
        self.__convert_graph_format(
            input_format=GraphFormat.ADJACENCY_MATRIX_BINARY, output_format=GraphFormat.BITMASK_IN
        )

        return self.__bitmask_in

    @property
    def adjacency_matrix_colors(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the adjacency matrix format with color numbers.
        """

        # Return the stored array if available.
        if self.__adjacency_matrix_colors is not None:
            return self.__adjacency_matrix_colors

        # Compute from the adjacency matrix format with binary slices or the bitmask formats.
        if (
            self.__adjacency_matrix_binary is not None
            or self.__bitmask_out is not None
            or self.__bitmask_in is not None
        ):
            self.adjacency_matrix_binary
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
                output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
            )
        # Compute from the flattened row-major format with color numbers.
        elif self.__flattened_row_major_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
            )
        # Compute from the flattened clockwise format with color numbers.
        elif self.__flattened_clockwise_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
            )
        # Compute from the flattened row-major format with binary slices.
        elif self.__flattened_row_major_binary is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
            )
        # Compute from the flattened clockwise format with binary slices.
        else:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
            )

        return self.__adjacency_matrix_colors

    @property
    def adjacency_matrix_binary(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the adjacency matrix format with binary slices.
        """

        # Return the stored array if available.
        if self.__adjacency_matrix_binary is not None:
            return self.__adjacency_matrix_binary

        # Compute from the bitmask format for the out-neighborhoods.
        if self.__bitmask_out is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.BITMASK_OUT,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
        # Compute from the bitmask format for the in-neighborhoods.
        elif self.__bitmask_in is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.BITMASK_IN,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
        # Compute from the adjacency matrix format with color numbers.
        elif self.__adjacency_matrix_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
        # Compute from the flattened formats with binary slices.
        elif self.__flattened_row_major_binary is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
        elif self.__flattened_clockwise_binary is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
        # Compute from the flattened formats with color numbers, if needed.
        else:
            if self.__flattened_row_major_colors is not None:
                self.__convert_graph_format(
                    input_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
                    output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                )
            else:
                self.__convert_graph_format(
                    input_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
                    output_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                )

            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )

        return self.__adjacency_matrix_binary

    @property
    def flattened_row_major_colors(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the flattened row-major format with color numbers.
        """

        # Return the stored array if available.
        if self.__flattened_row_major_colors is not None:
            return self.__flattened_row_major_colors

        # Compute from the flattened row-major format with binary slices, if available.
        if self.__flattened_row_major_binary is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
            )
        # Otherwise, compute via the adjacency matrix format with color numbers.
        else:
            self.adjacency_matrix_colors
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
            )

        return self.__flattened_row_major_colors

    @property
    def flattened_row_major_binary(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the flattened row-major format with binary slices.
        """

        # Return the stored array if available.
        if self.__flattened_row_major_binary is not None:
            return self.__flattened_row_major_binary

        # Compute from the flattened row-major format with color numbers, if available.
        if self.__flattened_row_major_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
            )
        # Compute from the adjacency matrix format with binary slices or the bitmask formats.
        elif (
            self.__adjacency_matrix_binary is not None
            or self.__bitmask_out is not None
            or self.__bitmask_in is not None
        ):
            self.adjacency_matrix_binary
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
            )
        # Compute from the adjacency matrix format with color numbers, if available.
        elif self.__adjacency_matrix_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_COLORS,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
            )
        # Otherwise, compute via the flattened clockwise format with binary slices.
        else:
            self.flattened_clockwise_binary
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
                output_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
            )

        return self.__flattened_row_major_binary

    @property
    def flattened_clockwise_colors(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the flattened clockwise format with color numbers.
        """

        # Return the stored array if available.
        if self.__flattened_clockwise_colors is not None:
            return self.__flattened_clockwise_colors

        # Compute from the flattened clockwise format with binary slices, if available.
        if self.__flattened_clockwise_binary is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
            )
        # Otherwise, compute via the adjacency matrix format with color numbers.
        else:
            self.adjacency_matrix_colors
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
            )

        return self.__flattened_clockwise_colors

    @property
    def flattened_clockwise_binary(self) -> np.ndarray:
        """
        This property returns the `numpy.ndarray` of type `numpy.uint8` representing the given
        graph or batch of graphs in the flattened clockwise format with binary slices.
        """

        # Return the stored array if available.
        if self.__flattened_clockwise_binary is not None:
            return self.__flattened_clockwise_binary

        # Compute from the flattened clockwise format with color numbers, if available.
        if self.__flattened_clockwise_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
            )
        # Compute from the adjacency matrix format with binary slices or the bitmask formats.
        elif (
            self.__adjacency_matrix_binary is not None
            or self.__bitmask_out is not None
            or self.__bitmask_in is not None
        ):
            self.adjacency_matrix_binary
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
            )
        # Compute from the adjacency matrix format with color numbers, if available.
        elif self.__adjacency_matrix_colors is not None:
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_COLORS,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_CLOCKWISE_COLORS,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
            )
        # Otherwise, compute via the flattened row-major format with binary slices.
        else:
            self.flattened_row_major_binary
            self.__convert_graph_format(
                input_format=GraphFormat.FLATTENED_ROW_MAJOR_BINARY,
                output_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
            )
            self.__convert_graph_format(
                input_format=GraphFormat.ADJACENCY_MATRIX_BINARY,
                output_format=GraphFormat.FLATTENED_CLOCKWISE_BINARY,
            )

        return self.__flattened_clockwise_binary

    def __getitem__(self, index: int) -> Graph:
        """
        This method returns the graph at the given index if this object models a batch of graphs.
        If this object models a single graph, it raises an `IndexError`.

        :param index: A nonnegative `int` below the batch size specifying which graph to extract.

        :return: The graph at the given index, as a `Graph` object.
        """

        if self.__batch_size is None:
            raise IndexError("Cannot index a single graph.")

        # Extract all computed graph format representations for the selected graph.
        bitmask_out = None if self.__bitmask_out is None else self.__bitmask_out[index]
        bitmask_in = None if self.__bitmask_in is None else self.__bitmask_in[index]
        adjacency_matrix_colors = (
            None
            if self.__adjacency_matrix_colors is None
            else self.__adjacency_matrix_colors[index]
        )
        adjacency_matrix_binary = (
            None
            if self.__adjacency_matrix_binary is None
            else self.__adjacency_matrix_binary[index]
        )
        flattened_row_major_colors = (
            None
            if self.__flattened_row_major_colors is None
            else self.__flattened_row_major_colors[index]
        )
        flattened_row_major_binary = (
            None
            if self.__flattened_row_major_binary is None
            else self.__flattened_row_major_binary[index]
        )
        flattened_clockwise_colors = (
            None
            if self.__flattened_clockwise_colors is None
            else self.__flattened_clockwise_colors[index]
        )
        flattened_clockwise_binary = (
            None
            if self.__flattened_clockwise_binary is None
            else self.__flattened_clockwise_binary[index]
        )

        # Return a new `Graph` object representing the selected graph.
        return Graph(
            edge_colors=self.__edge_colors,
            is_directed=self.__is_directed,
            allow_loops=self.__allow_loops,
            bitmask_out=bitmask_out,
            bitmask_in=bitmask_in,
            adjacency_matrix_colors=adjacency_matrix_colors,
            adjacency_matrix_binary=adjacency_matrix_binary,
            flattened_row_major_colors=flattened_row_major_colors,
            flattened_row_major_binary=flattened_row_major_binary,
            flattened_clockwise_colors=flattened_clockwise_colors,
            flattened_clockwise_binary=flattened_clockwise_binary,
        )



"""
The `rlgt.agents` package defines the `GraphAgent` abstract base class, which formalizes a
reinforcement learning agent for graph theory applications. It also provides three concrete agent
classes that inherit from `GraphAgent` and implement the Deep-Cross Entropy, REINFORCE and Proximal
Policy Optimization (PPO) methods, all based on ``PyTorch``. In addition, the package includes
several auxiliary classes for defining random action mechanisms.

This package can only be used if the optional ``agents`` extra dependencies are installed.
"""

try:
    import torch  # noqa: F401
except ModuleNotFoundError:
    print("To use the 'rlgt.agents' package, please install the 'agents' extra dependencies.")
    exit(1)

from .deep_cross_entropy_agent import DeepCrossEntropyAgent
from .graph_agent import GraphAgent
from .ppo_agent import PPOAgent
from .random_action_mechanisms import (
    ConstantRandomActionMechanism,
    ExponentialRandomActionMechanism,
    NoRandomActionMechanism,
    RandomActionMechanism,
)
from .reinforce_agent import ReinforceAgent


__all__ = [
    "DeepCrossEntropyAgent",
    "GraphAgent",
    "PPOAgent",
    "ConstantRandomActionMechanism",
    "ExponentialRandomActionMechanism",
    "NoRandomActionMechanism",
    "RandomActionMechanism",
    "ReinforceAgent",
]



"""
This ``Python`` module defines three linear reinforcement learning environments that inherit from
the `GraphEnvironment` class and model graph building games where the edges (resp. arcs) are all
initially either uncolored, or fully colored in some predetermined manner, and are then properly
(re)colored one by one, either in the row-major order or the clockwise order.
"""

from typing import Optional

import numpy as np

from ..graphs.graph import Graph
from ..graphs.graph_formats import ColorRepresentation, FlattenedOrdering, GraphFormat
from ..graphs.special_graphs import MonochromaticGraph
from ..graphs.utils import graph_order_to_flattened_length
from .graph_environment import (
    EpisodeStatus,
    GraphEnvironment,
    GraphInvariant,
    GraphInvariantDiff,
)
from .graph_generators import GraphGenerator, create_fixed_graph_generator


class LinearBuildEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game in which
    the edges (resp. arcs) are initially uncolored and are then properly colored one by one, either
    in row-major order or clockwise order. Users can configure the graph order, the number of
    proper edge colors, whether the graphs are directed or undirected, and whether loops are
    allowed. Edges (resp. arcs) are properly colored according to the reconstruction from a
    flattened format. If loops are not allowed, the corresponding edges are ignored; if the graphs
    are undirected, only the upper-triangular part of the adjacency matrix is considered (including
    or excluding the diagonal depending on whether loops are allowed), and each edge is effectively
    colored in both directions simultaneously.

    The RL tasks in this environment are episodic. The episode length equals the flattened length,
    i.e., the number of entries in any of the two flattened formats with color numbers. This number
    depends on the graph order, on whether the graphs are directed, and on whether loops are
    allowed.

    Each state is represented by a binary `numpy.ndarray` vector of type `numpy.uint8` and length
    ``edge_colors * flattened_length``, where ``edge_colors`` is the configured number of proper
    edge colors and ``flattened_length`` is the flattened length of the graph. In the state
    vectors, the first ``flattened_length`` bits indicate which edges (resp. arcs) have been
    colored with color 1, the next ``flattened_length`` bits indicate which edges (resp. arcs) have
    been colored with color 2, and this pattern continues for each color up to ``edge_colors - 1``.
    The final ``flattened_length`` bits form a one-hot encoding of the next edge (resp. arc) to be
    properly colored. A value of 1 at a given position indicates which edge (resp. arc) is to be
    colored next, while all zeros signify a terminal state in which all edges (resp. arcs) have
    been properly colored.

    Each action is represented by a `numpy.int32` integer between 0 and ``edge_colors - 1``, which
    specifies the color to assign to the next edge (resp. arc) in the selected order.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _edge_colors: The number of proper edge colors in the graphs to be constructed, given as
        a positive `int` that is at least 2.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying whether
        the edges (resp. arcs) are colored in row-major or clockwise order.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed, which also equals the total number of steps needed to reach a terminal state.
    :ivar _state_length: A positive `int` equal to ``_edge_colors * _flattened_length``, i.e., the
        length of each state vector.
    :ivar _step_count: Either `None` or a nonnegative `int` between 0 and `_flattened_length`
        indicating the index of the next edge (resp. arc) to properly color according to
        `_flattened_ordering`. When `_step_count` equals `_flattened_length`, the state is
        terminal. This attribute is initially `None` and updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `LinearBuildEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration indicating
            whether the edges (resp. arcs) should be properly colored in the row-major or clockwise
            order. The default value is `FlattenedOrdering.ROW_MAJOR`.
        :param edge_colors: A positive `int` (not below 2) specifying the number of proper edge
            colors in the graphs to be constructed. The default value is 2.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._edge_colors: int = edge_colors
        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._flattened_ordering: FlattenedOrdering = flattened_ordering
        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )
        self._state_length: int = self._edge_colors * self._flattened_length

        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._state_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        return self._edge_colors

    @property
    def action_mask(self) -> Optional[np.ndarray]:  # pragma: no cover
        return None

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._flattened_length

    @property
    def is_continuing(self) -> bool:
        return False

    def _initialize_batch(self, batch_size: int) -> None:
        self._state_batch = np.zeros((batch_size, self._state_length), dtype=np.uint8)
        # The zeroth edge (resp. arc) should be properly colored first.
        self._state_batch[:, -self._flattened_length] = 1
        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        # If the graphs have only two proper edge colors, then the transition can easily be done as
        # follows.
        if self._edge_colors == 2:
            self._state_batch[:, self._step_count] = action_batch
        # Otherwise, the next trick should be used.
        else:
            view = self._state_batch.reshape(-1, self._edge_colors, self._flattened_length)
            rows = np.arange(self._state_batch.shape[0])
            view[rows, action_batch - 1, self._step_count] = 1

        # Set the current edge (resp. arc) position flags to zero and increment the ``_step_count``
        # attribute.
        self._state_batch[:, -self._flattened_length + self._step_count] = 0
        self._step_count += 1

        if self._step_count < self._flattened_length:
            # Set the new current edge (resp. arc) position flags to one.
            self._state_batch[:, -self._flattened_length + self._step_count] = 1
        else:
            self._status = EpisodeStatus.TERMINATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        temp = state_batch.reshape(-1, self._edge_colors, self._flattened_length)

        # If all the underlying graphs are fully colored, then extract them in a reduced flattened
        # format with binary slices.
        if np.all(temp[:, -1, :] == 0):
            return Graph.from_flattened(
                flattened=temp[:, :-1, :].copy(),
                flattened_ordering=self._flattened_ordering,
                color_representation=ColorRepresentation.BINARY_SLICES,
                edge_colors=self._edge_colors,
                is_directed=self._is_directed,
                allow_loops=self._allow_loops,
            )

        # Otherwise, extract the underlying graphs in a standard (non-reduced) flattened format
        # with binary slices. First, settle the slices that correspond to the colors 1, 2, ...,
        # k - 1.
        result = np.empty(temp.shape, dtype=np.uint8)
        result[:, 1:, :] = temp[:, :-1, :]

        # Now, settle the slice that corresponds to color 0.
        uncolored_mask = np.maximum.accumulate(temp[:, -1, :], axis=1)
        result[:, 0, :] = 1 - uncolored_mask - np.sum(result[:, 1:, :], axis=1)

        return Graph.from_flattened(
            flattened=result,
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            edge_colors=self._edge_colors,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )


class LinearSetEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game in which
    the edges (resp. arcs) are initially fully colored and are then properly recolored one by one,
    either in row-major order or clockwise order. Users can configure the graph order, the number
    of proper edge colors, whether the graphs are directed or undirected, and whether loops are
    allowed. Additionally, the user can configure the mechanism that controls how the initial fully
    colored graphs are generated, which may be deterministic or nondeterministic. Edges (resp.
    arcs) are recolored according to the reconstruction from a flattened format. If loops are not
    allowed, the corresponding edges are ignored; if the graphs are undirected, only the
    upper-triangular part of the adjacency matrix is considered (including or excluding the
    diagonal depending on whether loops are allowed), and each edge is effectively recolored in
    both directions simultaneously.

    The RL tasks in this environment are episodic. The episode length equals the flattened length,
    i.e., the number of entries in any of the two flattened formats with color numbers. This number
    depends on the graph order, on whether the graphs are directed, and on whether loops are
    allowed.

    Each state is represented by a binary `numpy.ndarray` vector of type `numpy.uint8` and length
    ``edge_colors * flattened_length``, where ``edge_colors`` is the configured number of proper
    edge colors and ``flattened_length`` is the flattened length of the graph. In the state
    vectors, the first ``flattened_length`` bits indicate which edges (resp. arcs) currently have
    color 1, the next ``flattened_length`` bits indicate which edges (resp. arcs) have color 2, and
    this pattern continues up to ``edge_colors - 1``. The final ``flattened_length`` bits form a
    one-hot encoding of the next edge (resp. arc) to be properly recolored. A value of 1 at a given
    position indicates which edge (resp. arc) is to be recolored next, while all zeros signify a
    terminal state in which all edges (resp. arcs) have been properly recolored.

    Each action is represented by a `numpy.int32` integer between 0 and ``edge_colors - 1``, which
    specifies the color to assign to the next edge (resp. arc) in the selected order.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _edge_colors: The number of proper edge colors in the graphs to be constructed, given as
        a positive `int` that is at least 2.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying whether
        the edges (resp. arcs) are recolored in row-major or clockwise order.
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed, which also equals the total number of steps needed to reach a terminal state.
    :ivar _state_length: A positive `int` equal to ``_edge_colors * _flattened_length``, i.e., the
        length of each state vector.
    :ivar _step_count: Either `None` or a nonnegative `int` between 0 and `_flattened_length`
        indicating the index of the next edge (resp. arc) to properly recolored according to
        `_flattened_ordering`. When `_step_count` equals `_flattened_length`, the state is
        terminal. This attribute is initially `None` and updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `LinearSetEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration indicating
            whether the edges (resp. arcs) should be properly recolored in the row-major or
            clockwise order. The default value is `FlattenedOrdering.ROW_MAJOR`.
        :param edge_colors: A positive `int` (not below 2) specifying the number of proper edge
            colors in the graphs to be constructed. The default value is 2.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._edge_colors: int = edge_colors
        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )
        self._state_length: int = self._edge_colors * self._flattened_length

        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._state_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        return self._edge_colors

    @property
    def action_mask(self) -> Optional[np.ndarray]:  # pragma: no cover
        return None

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._flattened_length

    @property
    def is_continuing(self) -> bool:
        return False

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        # Initialize the state vectors using the generated underlying fully colored graphs.
        temp = format_representation[:, -self._edge_colors + 1 :, :].reshape(batch_size, -1)
        self._state_batch = np.zeros((batch_size, self._state_length), dtype=np.uint8)
        self._state_batch[:, : -self._flattened_length] = temp
        # The zeroth edge (resp. arc) should be potentially flipped first.
        self._state_batch[:, -self._flattened_length] = 1

        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        # If the graphs have only two proper edge colors, then the transition can easily be done as
        # follows.
        if self._edge_colors == 2:
            self._state_batch[:, self._step_count] = action_batch
        # Otherwise, the next trick should be used.
        else:
            view = self._state_batch.reshape(-1, self._edge_colors, self._flattened_length)
            view[:, :, self._step_count] = 0
            rows = np.arange(self._state_batch.shape[0])
            view[rows, action_batch - 1, self._step_count] = 1

        # Set the current edge (resp. arc) position flags to zero and increment the ``_step_count``
        # attribute.
        self._state_batch[:, -self._flattened_length + self._step_count] = 0
        self._step_count += 1

        if self._step_count < self._flattened_length:
            # Set the new current edge (resp. arc) position flags to one.
            self._state_batch[:, -self._flattened_length + self._step_count] = 1
        else:
            self._status = EpisodeStatus.TERMINATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        result = state_batch.reshape(-1, self._edge_colors, self._flattened_length)[
            :, :-1, :
        ].copy()

        return Graph.from_flattened(
            flattened=result,
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            edge_colors=self._edge_colors,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )


class LinearFlipEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game for
    constructing 2-edge-colored looped complete graphs. The edges (resp. arcs) are initially fully
    colored in some manner and can then be potentially flipped one by one in either row-major or
    clockwise order. Users can configure the graph order, whether the graphs are directed or
    undirected, whether loops are allowed, and the mechanism used to generate the initial fully
    colored graphs, which may be deterministic or nondeterministic. The flipping of edges (resp.
    arcs) follows the reconstruction from a flattened format: if loops are not allowed, the
    corresponding edges are ignored, and if the graphs are undirected, only the upper-triangular
    part of the adjacency matrix is considered (including or excluding the diagonal depending on
    loop allowance), so each edge is flipped in both directions simultaneously. The user can select
    whether edges (resp. arcs) are flipped in row-major or clockwise order.

    The RL tasks in this environment are episodic, and the episode length equals the flattened
    length, i.e., the number of entries in either flattened format with color numbers. This length
    depends on the graph order, whether the graphs are directed, and whether loops are allowed.

    Each state is represented by a binary `numpy.ndarray` vector of type `numpy.uint8` and length
    ``2 * flattened_length``, where ``flattened_length`` is the flattened length of the graphs. In
    the state vectors, the first ``flattened_length`` bits indicate which edges (resp. arcs) are
    currently colored with color 1, while the final ``flattened_length`` bits form a one-hot
    encoding of the edge (resp. arc) to be potentially flipped next. A 1 at a given position
    identifies the next edge (resp. arc) to potentially flip, while all zeros indicate a terminal
    state in which all edges (resp. arcs) have been traversed.

    Each action is represented by a `numpy.int32` integer, either 0 or 1. The value 1 flips the
    color of the next edge (resp. arc), while 0 leaves it unchanged.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying whether
        the edges (resp. arcs) are potentially flipped in row-major or clockwise order.
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed, which also equals the total number of steps needed to reach a terminal state.
    :ivar _step_count: Either `None` or a nonnegative `int` between 0 and `_flattened_length`
        indicating the index of the next edge (resp. arc) to potentially flipped according to
        `_flattened_ordering`. When `_step_count` equals `_flattened_length`, the state is
        terminal. This attribute is initially `None` and updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `LinearFlipEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration indicating
            whether the edges (resp. arcs) should be potentially flipped in the row-major or
            clockwise order. The default value is `FlattenedOrdering.ROW_MAJOR`.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )

        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return 2 * self._flattened_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        return 2

    @property
    def action_mask(self) -> Optional[np.ndarray]:  # pragma: no cover
        return None

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._flattened_length

    @property
    def is_continuing(self) -> bool:
        return False

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        self._state_batch = np.zeros((batch_size, 2 * self._flattened_length), dtype=np.uint8)
        self._state_batch[:, : self._flattened_length] = format_representation[:, -1, :]
        # The zeroth edge (resp. arc) should be potentially flipped first.
        self._state_batch[:, self._flattened_length] = 1

        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        self._state_batch[:, self._step_count] ^= action_batch.astype(np.uint8)
        # Set the current edge (resp. arc) position flags to zero and increment the ``_step_count``
        # attribute.
        self._state_batch[:, self._flattened_length + self._step_count] = 0
        self._step_count += 1

        if self._step_count < self._flattened_length:
            # Set the new current edge (resp. arc) position flags to one.
            self._state_batch[:, self._flattened_length + self._step_count] = 1
        else:
            self._status = EpisodeStatus.TERMINATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        result = state_batch.reshape(-1, 2, self._flattened_length)[:, :-1, :].copy()

        return Graph.from_flattened(
            flattened=result,
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )



"""
This ``Python`` module defines two local reinforcement learning environments that inherit from the
`GraphEnvironment` class and model graph building games in which the edges (resp. arcs) are
initially fully colored in some predetermined manner. In these environments, the agent moves
between vertices according to a specified traversal strategy, thereby traversing existing edges
(resp. arcs) and properly recoloring them as part of the interaction process.
"""

from typing import Optional

import numpy as np

from ..graphs.graph import Graph
from ..graphs.graph_formats import ColorRepresentation, FlattenedOrdering, GraphFormat
from ..graphs.special_graphs import MonochromaticGraph
from ..graphs.utils import compute_edge_indices, graph_order_to_flattened_length
from .graph_environment import (
    EpisodeStatus,
    GraphEnvironment,
    GraphInvariant,
    GraphInvariantDiff,
)
from .graph_generators import GraphGenerator, create_fixed_graph_generator


class LocalSetEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game in which
    the edges (resp. arcs) are initially fully colored in some manner, and an agent moves between
    vertices, thereby traversing existing edges (resp. arcs) and properly recoloring them. More
    precisely, at each step the agent is located at a vertex and selects an edge incident to this
    vertex (resp. an arc starting at this vertex), traverses it, and moves to the other endpoint.
    During traversal, the selected edge (resp. arc) is properly recolored with a chosen color. The
    user can select the graph order, the number of proper edge colors, whether the graphs are
    directed or undirected, and whether loops are allowed. Additionally, the mechanism controlling
    how the initial fully colored graphs are generated can be configured and may be deterministic
    or nondeterministic. The user can also select the vertex at which the agent starts the
    recoloring procedure.

    The RL tasks in this environment are continuing, and the total number of actions to be
    performed, i.e., the episode length, is configurable.

    Each state is represented by a binary `numpy.ndarray` of type `numpy.uint8` and length
    ``(edge_colors - 1) * flattened_length + graph_order``, where ``edge_colors`` is the configured
    number of proper edge colors, ``graph_order`` is the configured graph order, and
    ``flattened_length`` is the flattened length of the graphs to be constructed. In the state
    vectors, the first ``flattened_length`` bits indicate which edges (resp. arcs) are currently of
    color 1, the next ``flattened_length`` bits indicate which edges (resp. arcs) are currently of
    color 2, and this pattern continues up to color ``edge_colors - 1``. The ordering of edges
    (resp. arcs) within these blocks is determined by the selected flattened ordering, which can be
    either row-major or clockwise as specified by the `FlattenedOrdering` enumeration. The final
    ``graph_order`` bits form a one-hot encoding that specifies the vertex at which the agent is
    currently located.

    Each action is represented by a `numpy.int32` integer between 0 and
    ``edge_colors * graph_order - 1``. If the action value is ``a``, then ``a % graph_order``
    determines the vertex to which the agent should move from its current location, while
    ``a // graph_order`` determines the color with which the traversed edge (resp. arc) is properly
    recolored. If loops are not allowed, then actions that would keep the agent at the current
    vertex are invalid.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _edge_colors: The number of proper edge colors in the graphs to be constructed, given as
        a positive `int` that is at least 2.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _graph_order: A positive `int` that describes the order of the graphs to be constructed.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying the edge
        (resp. arc) ordering (row-major or clockwise).
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar starting_vertex: A nonnegative `int` below the configured graph order that determines
        the vertex at which the agent should start the recoloring procedure. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed.
    :ivar _state_length: A positive `int` that determines the length of each of the state vectors,
        i.e., the number ``(_edge_colors - 1) * _flattened_length + _graph_order``.
    :ivar _episode_length: A positive `int` specifying the episode length, i.e., the total number
        of actions in each episode.
    :ivar _current_vertices: Either `None` or a `numpy.ndarray` of type `numpy.int32` specifying
        the vertex where the agent is currently located in each episode run in parallel.
    :ivar _step_count: Either `None` or a nonnegative `int` counting how many actions have been
        executed in the current batch of episodes. When `_step_count` equals `_episode_length`, the
        episode has reached a final state. This attribute is updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        episode_length: Optional[int] = None,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        edge_colors: int = 2,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        starting_vertex: int = 0,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `LocalSetEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param episode_length: Either `None`, or a positive `int` specifying the number of actions
            in each episode. If `None`, the episode length defaults to the flattened length of the
            graphs to be constructed. The default value is `None`.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying
            whether the edges (resp. arcs) are ordered row-major or clockwise. The default value is
            `FlattenedOrdering.ROW_MAJOR`.
        :param edge_colors: A positive `int` (not below 2) specifying the number of proper edge
            colors in the graphs to be constructed. The default value is 2.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param starting_vertex: A nonnegative `int` strictly less than ``graph_order`` specifying
            the vertex at which the agent starts the recoloring procedure. The default value is 0.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._edge_colors: int = edge_colors
        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._graph_order: int = graph_order
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    edge_colors=edge_colors,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self.starting_vertex: int = starting_vertex

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )
        self._state_length: int = (
            self._edge_colors - 1
        ) * self._flattened_length + self._graph_order

        if episode_length is not None:
            self._episode_length: int = episode_length
        # If the ``episode_length`` argument is `None`, then the episode length should match the
        # ``_flattened_length`` attribute.
        else:
            self._episode_length: int = self._flattened_length

        self._current_vertices: Optional[np.ndarray] = None
        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._state_length

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        return self._edge_colors * self._graph_order

    @property
    def action_mask(self) -> Optional[np.ndarray]:
        if self._status is None or self._status != EpisodeStatus.IN_PROGRESS or self._allow_loops:
            return None

        result = np.ones(
            (self._state_batch.shape[0], self._edge_colors, self._graph_order), dtype=bool
        )
        rows = np.arange(self._state_batch.shape[0])
        result[rows, :, self._current_vertices] = False

        return result.reshape(self._state_batch.shape[0], -1)

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._episode_length

    @episode_length.setter
    def episode_length(self, episode_length: int):  # pragma: no cover
        """
        This setter allows the user to potentially reconfigure the episode length between two
        independent batches of episodes. It should not be used while a batch of episodes is
        currently in progress.

        :param episode_length: A positive `int` specifying the new episode length.
        """

        self._episode_length = episode_length

    @property
    def is_continuing(self) -> bool:
        return True

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        # Initialize the state vectors using the generated underlying fully colored graphs. Also,
        # make sure that the current vertex position flags are all set to the configured starting
        # vertex.
        self._state_batch = np.zeros((batch_size, self._state_length), dtype=np.uint8)
        self._state_batch[:, : -self._graph_order] = format_representation[
            :, -self._edge_colors + 1 :, :
        ].reshape(batch_size, -1)
        self._state_batch[:, -self._graph_order + self.starting_vertex] = 1

        self._current_vertices = np.full((batch_size,), self.starting_vertex, dtype=np.int32)
        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        """
        This method applies a batch of actions to the current batch of states and updates the
        `_state_batch` and `_status` attributes to reflect the resulting states and the updated
        batch status.

        :param action_batch: A one-dimensional `numpy.ndarray` of type `numpy.int32` containing the
            actions to be applied. The length of ``action_batch`` must match the number of states
            in ``_state_batch``.

        :note: If loops are not allowed and an action attempts to traverse a loop, a `RuntimeError`
            is raised.
        """

        new_vertices = action_batch % self._graph_order
        new_colors = action_batch // self._graph_order

        # If the agent attempts to traverse a nonexisting loop, a `RuntimeError` will be raised.
        if not self._allow_loops:
            if np.any(self._current_vertices == new_vertices):
                raise RuntimeError("Cannot traverse a nonexisting loop.")

        # Use the ``compute_edge_indices`` auxiliary function to compute the index of the edge
        # (resp. arc) that should be traversed in each of the episodes run in parallel.
        edge_indices = compute_edge_indices(
            graph_order=self._graph_order,
            starting_vertices=self._current_vertices,
            ending_vertices=new_vertices,
            flattened_ordering=self._flattened_ordering,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )

        rows = np.arange(self._state_batch.shape[0])

        # If the graphs have only two proper edge colors, then the transition can easily be done as
        # follows.
        if self._edge_colors == 2:
            self._state_batch[rows, edge_indices] = new_colors
        # Otherwise, the next trick should be used.
        else:
            temp = self._state_batch[:, : -self._graph_order].reshape(
                -1, self._edge_colors - 1, self._flattened_length
            )
            temp[rows, :, edge_indices] = 0
            temp[rows, new_colors - 1, edge_indices] = new_colors != 0
            self._state_batch[:, : -self._graph_order] = temp.reshape(temp.shape[0], -1)

        # Update the current vertex position flags.
        self._state_batch[rows, self._current_vertices - self._graph_order] = 0
        self._current_vertices = new_vertices
        self._state_batch[rows, self._current_vertices - self._graph_order] = 1

        self._step_count += 1
        if self._step_count >= self._episode_length:
            self._status = EpisodeStatus.TRUNCATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        result = (
            state_batch[:, : -self._graph_order]
            .reshape(-1, self._edge_colors - 1, self._flattened_length)
            .copy()
        )

        return Graph.from_flattened(
            flattened=result,
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            edge_colors=self._edge_colors,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )


class LocalFlipEnvironment(GraphEnvironment):
    """
    This class inherits from the `GraphEnvironment` class and models a graph building game used for
    constructing 2-edge-colored looped complete graphs in which the edges (resp. arcs) are
    initially fully colored in some manner, and the agent moves from one vertex to another, thereby
    traversing existing edges (resp. arcs) and potentially flipping them. More precisely, in each
    step, the agent is located at a vertex and selects an incident edge (resp. an outgoing arc),
    traverses it, and moves to the opposite endpoint. While traversing an edge (resp. arc), the
    agent may flip its proper edge color. The user can select the graph order, choose whether the
    graphs are directed or undirected, and specify whether loops are allowed. Additionally, the
    user can configure the mechanism that controls how the initial fully colored graphs are
    generated, which may be either deterministic or nondeterministic. The user can also select the
    vertex at which the agent starts the potential flipping procedure.

    The RL tasks in this environment are continuing, and the total number of actions to be
    performed, i.e., the episode length, is configurable.

    Each state is represented by a binary `numpy.ndarray` of type `numpy.uint8` and length
    ``flattened_length + graph_order``, where ``graph_order`` is the configured graph order and
    ``flattened_length`` is the flattened length of the graphs to be constructed. In the state
    vectors, the first ``flattened_length`` bits indicate which of the ``flattened_length`` edges
    (resp. arcs) are currently of color 1. The edges (resp. arcs) are ordered according to the
    selected flattened ordering, which may be either row-major or clockwise. The final
    ``graph_order`` bits form a one-hot encoding of the vertex at which the agent is currently
    located.

    The environment supports two action-handling modes. If ``flip_only`` is set to `False` (the
    default), each action is represented by a `numpy.int32` integer between 0 and
    ``2 * graph_order - 1``. For an action value ``a``, the quantity ``a % graph_order`` specifies
    the vertex to which the agent moves, while ``a // graph_order`` determines whether the
    traversed edge (resp. arc) is flipped. A value of 1 indicates that the proper edge color is
    changed, whereas 0 indicates that it remains unchanged. If ``flip_only`` is set to `True`,
    then every traversed edge (resp. arc) must be flipped. In this case, each action is a
    `numpy.int32` integer between 0 and ``graph_order - 1`` specifying the destination vertex, with
    the traversed edge (resp. arc) necessarily flipped. If loops are not allowed, then the agent
    cannot move to its current vertex.

    :ivar _state_batch: See the description of the `GraphEnvironment._state_batch` attribute.
    :ivar _status: See the description of the `GraphEnvironment._status` attribute.
    :ivar _is_directed: A `bool` indicating whether the graphs to be constructed are directed or
        undirected.
    :ivar _allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
        constructed.
    :ivar _graph_order: A positive `int` that describes the order of the graphs to be constructed.
    :ivar _flip_only: A `bool` indicating whether all traversed edges (resp. arcs) must be flipped.
    :ivar _flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying the edge
        (resp. arc) ordering (row-major or clockwise).
    :ivar initial_graph_generator: A `GraphGenerator` function that defines how the underlying
        fully colored graphs are generated for the initial states. This attribute may be
        reconfigured between independent batches of episodes.
    :ivar starting_vertex: A nonnegative `int` below the configured graph order that determines
        the vertex at which the agent should start the potential flipping procedure. This attribute
        may be reconfigured between independent batches of episodes.
    :ivar _flattened_length: A positive `int` equal to the flattened length of the graphs to be
        constructed.
    :ivar _episode_length: A positive `int` specifying the episode length, i.e., the total number
        of actions in each episode.
    :ivar _current_vertices: Either `None` or a `numpy.ndarray` of type `numpy.int32` specifying
        the vertex where the agent is currently located in each episode run in parallel.
    :ivar _step_count: Either `None` or a nonnegative `int` counting how many actions have been
        executed in the current batch of episodes. When `_step_count` equals `_episode_length`, the
        episode has reached a final state. This attribute is updated after each call to
        `GraphEnvironment.reset_batch` or `GraphEnvironment.step_batch`.
    """

    def __init__(
        self,
        graph_invariant: GraphInvariant,
        graph_order: int,
        episode_length: Optional[int] = None,
        flip_only: bool = False,
        flattened_ordering: FlattenedOrdering = FlattenedOrdering.ROW_MAJOR,
        is_directed: bool = False,
        allow_loops: bool = False,
        initial_graph_generator: Optional[GraphGenerator] = None,
        starting_vertex: int = 0,
        graph_invariant_diff: Optional[GraphInvariantDiff] = None,
        sparse_setting: bool = False,
    ):
        """
        This constructor initializes an instance of the `LocalFlipEnvironment` class.

        :param graph_invariant: A `GraphInvariant` function that computes the graph invariant
            values associated with a batch of underlying graphs. These values are the quantities to
            be maximized by the environment.
        :param graph_order: A positive `int` (not below 2) that represents the graph order of the
            graphs to be constructed.
        :param episode_length: Either `None`, or a positive `int` specifying the number of actions
            in each episode. If `None`, the episode length defaults to the flattened length of the
            graphs to be constructed. The default value is `None`.
        :param flip_only: A `bool` specifying whether the traversed edges (resp. arcs) must be
            flipped. The default value is `False`.
        :param flattened_ordering: An item of the `FlattenedOrdering` enumeration specifying
            whether the edges (resp. arcs) are ordered row-major or clockwise. The default value is
            `FlattenedOrdering.ROW_MAJOR`.
        :param is_directed: A `bool` indicating whether the graphs to be constructed are directed.
            The default value is `False`.
        :param allow_loops: A `bool` indicating whether loops are allowed in the graphs to be
            constructed. The default value is `False`.
        :param initial_graph_generator: Either `None` or a `GraphGenerator` function that
            determines how the initial fully colored graphs are generated for the batch of initial
            states. If `None`, all edges (resp. arcs) in all graphs are initially colored with
            color 0. The default value is `None`.
        :param starting_vertex: A nonnegative `int` strictly less than ``graph_order`` specifying
            the vertex at which the agent starts the potential flipping procedure. The default
            value is 0.
        :param graph_invariant_diff: Either `None`, indicating that graph invariant values are
            always computed directly using ``graph_invariant``, or a `GraphInvariantDiff` function
            that computes element-wise differences of the graph invariant values when the
            environment transitions from one batch of underlying graphs to another. The default
            value is `None`.
        :param sparse_setting: A `bool` indicating whether the sparse setting is enabled. If set to
            `True`, the graph invariant values are computed only for the final batch of actions.
            Otherwise, the graph invariant values are computed after every batch of actions. The
            default value is `False`.
        """

        super().__init__(
            graph_invariant=graph_invariant,
            graph_invariant_diff=graph_invariant_diff,
            sparse_setting=sparse_setting,
        )

        self._is_directed: bool = is_directed
        self._allow_loops: bool = allow_loops
        self._graph_order: int = graph_order
        self._flip_only: bool = flip_only
        self._flattened_ordering: FlattenedOrdering = flattened_ordering

        if initial_graph_generator is not None:
            self.initial_graph_generator: GraphGenerator = initial_graph_generator
        else:
            # By default, all the edges (resp. arcs) in all the graphs should be colored with color
            # 0.
            graph_format = (
                GraphFormat.FLATTENED_ROW_MAJOR_BINARY
                if flattened_ordering == FlattenedOrdering.ROW_MAJOR
                else GraphFormat.FLATTENED_CLOCKWISE_BINARY
            )
            self.initial_graph_generator: GraphGenerator = create_fixed_graph_generator(
                fixed_graph=MonochromaticGraph(
                    graph_formats={graph_format},
                    graph_order=graph_order,
                    is_directed=is_directed,
                    allow_loops=allow_loops,
                ),
                graph_format=graph_format,
            )

        self.starting_vertex: int = starting_vertex

        self._flattened_length: int = graph_order_to_flattened_length(
            graph_order=graph_order,
            is_directed=is_directed,
            allow_loops=allow_loops,
        )

        if episode_length is not None:
            self.episode_length: int = episode_length
        # If the ``episode_length`` argument is `None`, then the episode length should match the
        # ``_flattened_length`` attribute.
        else:
            self.episode_length: int = self._flattened_length

        self._current_vertices: Optional[np.ndarray] = None
        self._step_count: Optional[int] = None

    @property
    def state_length(self) -> int:
        return self._flattened_length + self._graph_order

    @property
    def state_dtype(self) -> np.dtype:
        return np.uint8

    @property
    def action_number(self) -> int:  # pragma: no cover
        if self._flip_only:
            return self._graph_order
        else:
            return 2 * self._graph_order

    @property
    def action_mask(self) -> Optional[np.ndarray]:
        if self._status is None or self._status != EpisodeStatus.IN_PROGRESS or self._allow_loops:
            return None

        rows = np.arange(self._state_batch.shape[0])

        if self._flip_only:
            result = np.ones((self._state_batch.shape[0], self._graph_order), dtype=bool)
            result[rows, self._current_vertices] = False

            return result

        result = np.ones((self._state_batch.shape[0], 2 * self._graph_order), dtype=bool)
        result[rows, self._current_vertices] = False
        result[rows, self._graph_order + self._current_vertices] = False

        return result

    @property
    def episode_length(self) -> int:  # pragma: no cover
        return self._episode_length

    @episode_length.setter
    def episode_length(self, episode_length: int):  # pragma: no cover
        """
        This setter allows the user to potentially reconfigure the episode length between two
        independent batches of episodes. It should not be used while a batch of episodes is
        currently in progress.

        :param episode_length: A positive `int` specifying the new episode length.
        """

        self._episode_length = episode_length

    @property
    def is_continuing(self) -> bool:
        return True

    def _initialize_batch(self, batch_size: int) -> None:
        # Use the ``initial_graph_generator`` to generate the initial underlying fully colored
        # graphs.
        initial_graph_batch = self.initial_graph_generator(batch_size=batch_size)
        if self._flattened_ordering == FlattenedOrdering.ROW_MAJOR:
            format_representation = initial_graph_batch.flattened_row_major_binary
        else:
            format_representation = initial_graph_batch.flattened_clockwise_binary

        # Initialize the state vectors using the generated underlying fully colored graphs. Also,
        # make sure that the current vertex position flags are all set to the configured starting
        # vertex.
        self._state_batch = np.zeros(
            (batch_size, self._flattened_length + self._graph_order), dtype=np.uint8
        )
        self._state_batch[:, : self._flattened_length] = format_representation[:, -1, :]
        self._state_batch[:, self._flattened_length + self.starting_vertex] = 1

        self._current_vertices = np.full((batch_size,), self.starting_vertex, dtype=np.int32)
        self._status = EpisodeStatus.IN_PROGRESS
        self._step_count = 0

    def _transition_batch(self, action_batch: np.ndarray) -> None:
        """
        This method applies a batch of actions to the current batch of states and updates the
        `_state_batch` and `_status` attributes to reflect the resulting states and the updated
        batch status.

        :param action_batch: A one-dimensional `numpy.ndarray` of type `numpy.int32` containing the
            actions to be applied. The length of ``action_batch`` must match the number of states
            in ``_state_batch``.

        :note: If loops are not allowed and an action attempts to traverse a loop, a `RuntimeError`
            is raised.
        """

        new_vertices = action_batch % self._graph_order

        # If the agent attempts to traverse a nonexisting loop, a `RuntimeError` will be raised.
        if not self._allow_loops:
            if np.any(self._current_vertices == new_vertices):
                raise RuntimeError("Cannot traverse a nonexisting loop.")

        # Use the ``compute_edge_indices`` auxiliary function to compute the index of the edge
        # (resp. arc) that should be traversed in each of the episodes run in parallel.
        edge_indices = compute_edge_indices(
            graph_order=self._graph_order,
            starting_vertices=self._current_vertices,
            ending_vertices=new_vertices,
            flattened_ordering=self._flattened_ordering,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )

        rows = np.arange(self._state_batch.shape[0], dtype=np.int32)
        # If the ``_flip_only`` attribute is `True`, then all the traversed edges (resp. arcs) must
        # be flipped.
        if self._flip_only:
            self._state_batch[rows, edge_indices] ^= 1
        # Otherwise, each traversed edge (resp. arc) is either flipped or not flipped, depending on
        # the second entry from the corresponding action, which is binary.
        else:
            to_flip_or_not_to_flip = action_batch // self._graph_order
            self._state_batch[rows, edge_indices] ^= to_flip_or_not_to_flip.astype(np.uint8)

        # Update the current vertex position flags.
        self._state_batch[rows, self._flattened_length + self._current_vertices] = 0
        self._current_vertices = new_vertices
        self._state_batch[rows, self._flattened_length + self._current_vertices] = 1

        self._step_count += 1
        if self._step_count >= self.episode_length:
            self._status = EpisodeStatus.TRUNCATED

    def state_batch_to_graph_batch(self, state_batch: np.ndarray) -> Graph:
        result = (
            state_batch[:, : self._flattened_length].reshape(-1, 1, self._flattened_length).copy()
        )

        return Graph.from_flattened(
            flattened=result,
            flattened_ordering=self._flattened_ordering,
            color_representation=ColorRepresentation.BINARY_SLICES,
            is_directed=self._is_directed,
            allow_loops=self._allow_loops,
        )



"""
This ``Python`` module contains the `PPOAgent` class, which implements a reinforcement learning
agent for graph theory applications using the Proximal Policy Optimization (PPO) method with
``PyTorch``.
"""

from typing import Callable, Optional

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Categorical

from ..environments.graph_environment import EpisodeStatus, GraphEnvironment
from ..graphs.graph import Graph
from .graph_agent import GraphAgent
from .random_action_mechanisms import NoRandomActionMechanism, RandomActionMechanism


class PPOAgent(GraphAgent):
    """
    This class encapsulates a reinforcement learning agent for graph theory applications using the
    ``PyTorch``-based Proximal Policy Optimization (PPO) method. The agent operates on a
    configurable environment given as a `GraphEnvironment` object. In each iteration of the
    learning process, the agent generates a predetermined number of graphs by playing the graph
    building game defined by the environment and computes the graph invariant values and all
    discounted returns for each episode run in parallel. Here, while computing a discounted return,
    a reward is considered to be the increase between two consecutive graph invariant values. The
    agent uses an actor-critic architecture, with a `torch.nn.Module` model (policy network)
    applied to compute the probability of selecting each action in each step of every episode, and
    another `torch.nn.Module` model (value network) applied to estimate the value that quantifies
    the desirability of each state. Afterwards, the log probabilities and discounted returns of a
    subset of top-performing episodes are used to train both models according to the PPO algorithm.
    The training is performed over a configurable number of epochs. This completes one iteration of
    the learning process. The user provides both models, configures the optimizer, sets the
    discount factor, selects the number of epochs to be executed, configures the clamping epsilon
    coefficient and the value loss coefficient, and optionally provides a random action mechanism.
    When a random action occurs, it is selected uniformly among all actions available in the
    current state.

    :ivar _environment: A `GraphEnvironment` object defining the extremal problem and providing the
        graph building game used to construct all the graphs.
    :ivar _policy_network: A `torch.nn.Module` object predicting the action probabilities for each
        step in each episode.
    :ivar _value_network: A `torch.nn.Module` object predicting the state values for each step in
        each episode.
    :ivar _optimizer: A `torch.optim.Optimizer` object that updates the parameters of both models.
    :ivar _value_loss_function: A function implementing the MSE loss used for training the value
        network.
    :ivar _device: A `torch.device` object indicating the device where the models reside.
    :ivar _candidates_count: A positive `int` specifying the number of graphs constructed per
        iteration, i.e., the number of episodes run in parallel.
    :ivar _elite_count: A positive `int` specifying the number of top-performing episodes used to
        train the model in each iteration, or `None` if all the episodes should be used.
    :ivar _discount_factor: A `float` from the interval [0, 1] representing the discount factor to
        be used while computing the discounted returns.
    :ivar _epochs_count: A positive `int` specifying the number of epochs executed in each learning
        iteration of the PPO method.
    :ivar _clamp_epsilon: A `float` from the interval [0, 1] representing the clamping epsilon
        coefficient used while computing the policy loss in each epoch of a learning iteration.
    :ivar _value_loss_coef: A positive `float` representing the coefficient that scales the value
        loss while computing the total loss.
    :ivar _random_action_mechanism: A `RandomActionMechanism` object that determines the
        probability of executing a random action. When a random action is selected, it is sampled
        uniformly among all available actions in the current state.
    :ivar _random_generator: A `numpy.random.Generator` object used for all probabilistic
        decisions.
    :ivar _step_count: A nonnegative `int` representing the number of executed iterations, or
        `None` if the agent has not been initialized.
    :ivar _best_score: A `float` representing the best achieved value for the graph invariant, or
        `None` if the agent has not been initialized.
    :ivar _best_graph: A `Graph` object representing a graph attaining the best achieved value for
        the graph invariant, or `None` if the agent has not been initialized or no iterations have
        been executed.
    :ivar _population_states: Either `None` if uninitialized, or a `numpy.ndarray` storing all
        states during each episode trajectory. Its shape is ``(episode_length + 1,
        candidates_count, state_length)``, where ``episode_length`` is the episode length of the RL
        environment, ``state_length`` is the length of the state vectors, and ``candidates_count``
        is the number of episodes executed in parallel. The first dimension corresponds to the
        state trajectory within an episode, the second to the executed episodes, and the third to
        the state vector entries.
    :ivar _population_actions: Either `None` if uninitialized, or a `numpy.ndarray` of type
        `numpy.int32` storing all actions during each episode trajectory. Its shape is
        ``(episode_length, candidates_count)``, where the first dimension corresponds to the action
        trajectory within an episode and the second to the executed episodes. The episode order
        matches `_population_states`.
    :ivar _population_returns: Either `None` if uninitialized, or a `numpy.ndarray` of type
        `numpy.float32` storing the discounted returns for all executed episodes. Its shape is
        ``(episode_length, candidates_count)``, where the first dimension corresponds to the
        timestamps (actions) within an episode and the second to the executed episodes. The episode
        order matches `_population_states`.
    """

    def __init__(
        self,
        environment: GraphEnvironment,
        policy_network: nn.Module,
        value_network: nn.Module,
        optimizer: torch.optim.Optimizer,
        candidates_count: int = 200,
        elite_count: Optional[int] = None,
        discount_factor: float = 0.99,
        epochs_count: int = 4,
        clamp_epsilon: float = 0.2,
        value_loss_coef: float = 0.5,
        random_action_mechanism: RandomActionMechanism = NoRandomActionMechanism(),
        random_generator: Optional[np.random.Generator] = None,
    ):
        """
        This constructor initializes an instance of the `PPOAgent` class.

        :param environment: The RL environment defining the extremal problem and providing the
            graph building game, given as a `GraphEnvironment` object.
        :param policy_network: The policy network used to compute the probability of each action in
            each episode and step, given as a `torch.nn.Module` object.
        :param value_network: The value network used to compute the state value of each episode and
            step, given as a `torch.nn.Module` object. This value and policy networks must reside
            in the same device.
        :param optimizer: The optimizer responsible for updating the parameters of both models,
            given as a `torch.optim.Optimizer` object. The parameters of both ``policy_network``
            and ``value_network`` must be passed to it.
        :param candidates_count: A positive `int` specifying how many graphs are generated in each
            iteration by running the corresponding number of episodes in parallel. The default
            value is 200.
        :param elite_count: A positive `int` specifying how many episodes with the greatest graph
            invariant value are used to train the policy network and value network in each
            iteration of the learning process, or `None` to indicate that all executed episodes
            should be used. The default value is `None`.
        :param discount_factor: A `float` from the interval [0, 1] representing the discount factor
            to be used while computing the returns. The default value is 0.99.
        :param epochs_count: A positive `int` specifying the number of epochs executed in each
            learning iteration of the PPO method. The default value is 4.
        :param clamp_epsilon: A positive `float` from the interval [0, 1] representing the clamping
            epsilon coefficient used while computing the policy loss in each epoch of a learning
            iteration. The default value is 0.2.
        :param value_loss_coef: A positive `float` representing the coefficient that scales the
            value loss while computing the total loss. The default value is 0.5.
        :param random_action_mechanism: A `RandomActionMechanism` object that governs the
            probability of executing a random action in each step of the graph building game. When
            a random action is triggered, the agent ignores the action predicted by the policy
            network and instead selects an action uniformly at random among all available actions.
            By default, this is ``NoRandomActionMechanism()``, meaning that no random actions are
            ever executed.
        :param random_generator: Either `None`, or a `numpy.random.Generator` used for
            probabilistic decisions. If `None`, a default generator will be used. The default value
            is `None`.
        """

        self._environment: GraphEnvironment = environment
        # Enforce no sparse setting.
        self._environment.sparse_setting = False

        self._policy_network: nn.Module = policy_network
        self._value_network: nn.Module = value_network
        self._optimizer: torch.optim.Optimizer = optimizer
        self._value_loss_function: Callable = nn.MSELoss()

        # Infer the device from the policy network parameters. If no parameters exist, fall back to
        # the CPU.
        params = list(self._policy_network.parameters())
        self._device: torch.device = params[0].device if params else torch.device("cpu")

        self._candidates_count: int = candidates_count
        self._elite_count: Optional[int] = elite_count
        self._discount_factor: float = discount_factor
        self._epochs_count: int = epochs_count
        self._clamp_epsilon: float = clamp_epsilon
        self._value_loss_coef: float = value_loss_coef
        self._random_action_mechanism: RandomActionMechanism = random_action_mechanism

        # If the ``random_generator`` argument is `None`, then use a default `np.random.Generator`.
        if random_generator is None:
            random_generator = np.random.default_rng()
        self._random_generator: np.random.Generator = random_generator

        self._step_count: Optional[int] = None
        self._best_score: Optional[float] = None
        self._best_graph: Optional[Graph] = None

        self._population_states: Optional[np.ndarray] = None
        self._population_actions: Optional[np.ndarray] = None
        self._population_returns: Optional[np.ndarray] = None

    def reset(self) -> None:
        # Initialize the step count to 0, the best score to minus infinity, and the best graph to
        # `None`. Also, initialize the random action mechanism.
        self._step_count = 0
        self._best_score = float("-inf")
        self._best_graph = None
        self._random_action_mechanism.reset()

        # Initialize the population states, the population actions and the population returns to
        # the zero `np.ndarray` objects of the required shape and type.
        self._population_states = np.zeros(
            (
                self._environment.episode_length + 1,
                self._candidates_count,
                self._environment.state_length,
            ),
            dtype=self._environment.state_dtype,
        )
        self._population_actions = np.zeros(
            (self._environment.episode_length, self._candidates_count), dtype=np.int32
        )
        self._population_returns = np.zeros(
            (self._environment.episode_length, self._candidates_count), dtype=np.float32
        )

        # Reset the policy network parameters and the value network parameters.
        self._policy_network.apply(lambda m: getattr(m, "reset_parameters", lambda: None)())
        self._value_network.apply(lambda m: getattr(m, "reset_parameters", lambda: None)())

    def step(self) -> None:
        # Initialize a batch of episodes with the batch size ``_candidates_count`` and store the
        # starting states to the ``_population_states`` attribute.
        state_batch, current_scores, status = self._environment.reset_batch(
            batch_size=self._candidates_count
        )
        self._population_states[0, :, :] = state_batch

        # Initialize the new best score and new best graph.
        new_best_score = self._best_score
        new_best_graph = self._best_graph

        # If the RL environment is continuing, then determine the best graph and best score from
        # the starting timestamp, and update the corresponding variables if needed.
        if self._environment.is_continuing:
            timestamp_best = np.max(current_scores)
            if timestamp_best > new_best_score:
                new_best_score = timestamp_best
                best_index = np.argmax(current_scores)
                new_best_graph = self._environment.state_to_graph(state_batch[best_index, :])

        # Set the ``_population_returns`` attribute to all zeros. Also, initialize the
        # ``population_old_log_probs`` and ``population_old_values`` lists that store all the log
        # probabilities per timestamp and all the state values per timestamp, respectively.
        self._population_returns[:, :] = 0
        population_old_log_probs = []
        population_old_values = []

        # Set the episode action counter to 0 and use the random action mechanism to obtain the
        # random action probability.
        episode_action_count = 0
        random_action_probability = self._random_action_mechanism.random_action_probability

        # While the episodes are in progress...
        while status == EpisodeStatus.IN_PROGRESS:
            # Use the policy network to get the probability distribution for each action to be
            # selected for execution in each of the episodes run in parallel.
            state_batch_torch = torch.from_numpy(state_batch.astype(np.float32)).to(self._device)
            logits_batch_torch = self._policy_network(state_batch_torch)

            # Use the value network to get the state values and store them to the
            # ``population_old_values`` list.
            values_batch_torch = self._value_network(state_batch_torch)
            population_old_values.append(values_batch_torch.detach())

            # Make it impossible to execute an action that is not available for execution.
            action_mask = self._environment.action_mask
            if action_mask is not None:
                action_mask_torch = torch.from_numpy(action_mask).to(self._device)
                logits_batch_torch = logits_batch_torch.masked_fill(
                    ~action_mask_torch, float("-inf")
                )

            # Sample the actions according to the obtained probability distributions.
            distribution = Categorical(logits=logits_batch_torch)
            action_batch_torch = distribution.sample()
            action_batch = action_batch_torch.cpu().numpy()

            # Store the log probabilities to the ``population_old_log_probs`` list.
            population_old_log_probs.append(distribution.log_prob(action_batch_torch).detach())

            # Use the random action probability to decide whether each sampled action should be
            # replaced by a random action.
            random_mask = (
                self._random_generator.random(size=action_batch.shape[0])
                < random_action_probability
            )

            # Select each required random action among the actions available for execution using
            # the uniform probability distribution.
            if np.any(random_mask):
                # Settle the case where at least one action is not available for execution.
                if action_mask is not None:
                    probabilities_batch = action_mask[random_mask].astype(np.float32)
                    probabilities_batch /= probabilities_batch.sum(axis=1, keepdims=True)

                    action_batch[random_mask] = np.array(
                        [
                            self._random_generator.choice(
                                probabilities_batch.shape[1], p=probabilities
                            )
                            for probabilities in probabilities_batch
                        ],
                        dtype=np.int32,
                    )

                # Settle the case where all the actions are available for execution.
                else:
                    entry_count = np.count_nonzero(random_mask)
                    action_batch[random_mask] = self._random_generator.integers(
                        low=0,
                        high=self._environment.action_number,
                        size=entry_count,
                        dtype=np.int32,
                    )

            # Store the selected actions, execute them, and compute the batch of rewards.
            previous_scores = current_scores
            self._population_actions[episode_action_count, :] = action_batch
            state_batch, current_scores, status = self._environment.step_batch(action_batch)
            reward_batch = current_scores - previous_scores

            # Update the discounted returns.
            weights = self._discount_factor ** np.arange(episode_action_count, -1, -1)
            self._population_returns[: episode_action_count + 1, :] += np.outer(
                weights, reward_batch
            )

            # If the RL environment is continuing, or the final batch of actions has been executed,
            # then determine the best graph and best score from the current timestamp, and update
            # the corresponding variables if needed.
            if self._environment.is_continuing or status != EpisodeStatus.IN_PROGRESS:
                timestamp_best = np.max(current_scores)
                if timestamp_best > new_best_score:
                    new_best_score = timestamp_best
                    best_index = np.argmax(current_scores)
                    new_best_graph = self._environment.state_to_graph(state_batch[best_index, :])

            # Store the newly obtained states.
            episode_action_count += 1
            self._population_states[episode_action_count, :, :] = state_batch

        # Compute the mask that decides which executed episodes should be used to train the policy
        # network and the value network.
        elite_mask = np.zeros((self._candidates_count), dtype=bool)
        if self._elite_count is None:
            elite_mask[:] = True
        else:
            elite_mask[
                np.argpartition(current_scores, -self._elite_count)[-self._elite_count :]
            ] = True

        # Extract the (state, action) pairs from the elite executed episodes, i.e., the executed
        # episodes that should be used to train the policy network and the value network.
        elite_states = self._population_states[:-1, elite_mask, :].reshape(
            -1, self._environment.state_length
        )
        elite_actions = self._population_actions[:, elite_mask].reshape(-1)
        elite_states_torch = torch.from_numpy(elite_states.astype(np.float32)).to(self._device)
        elite_actions_torch = torch.from_numpy(elite_actions.astype(np.int64)).to(self._device)

        # Prepare the log probabilities from the elite executed episodes for training.
        elite_log_probs_torch = torch.cat(
            [
                population_old_log_probs[index][elite_mask]
                for index in range(self._environment.episode_length)
            ]
        ).reshape(-1)
        # Prepare the state values from the elite executed episodes for training.
        elite_values_torch = torch.cat(
            [
                population_old_values[index][elite_mask]
                for index in range(self._environment.episode_length)
            ]
        ).reshape(-1)

        # Prepare the discounted returns from the elite executed episodes for training. Also,
        # compute the advantages and normalize them.
        elite_returns = self._population_returns[:, elite_mask].reshape(-1)
        elite_returns_torch = torch.from_numpy(elite_returns.astype(np.float32)).to(self._device)

        elite_advantages_torch = elite_returns_torch - elite_values_torch
        elite_advantages_torch = (elite_advantages_torch - elite_advantages_torch.mean()) / (
            elite_advantages_torch.std() + 1e-8
        )

        # For each epoch...
        for _ in range(self._epochs_count):
            logits_batch_torch = self._policy_network(elite_states_torch)
            values_batch_torch = self._value_network(elite_states_torch).reshape(-1)

            # Compute the policy loss.
            new_log_probs_torch = Categorical(logits=logits_batch_torch).log_prob(
                elite_actions_torch
            )
            ratio = torch.exp(new_log_probs_torch - elite_log_probs_torch)
            loss_1 = ratio * elite_advantages_torch
            loss_2 = (
                torch.clamp(ratio, 1 - self._clamp_epsilon, 1 + self._clamp_epsilon)
                * elite_advantages_torch
            )
            policy_loss = torch.min(loss_1, loss_2).mean()

            # Compute the value loss and the total loss.
            value_loss = self._value_loss_function(values_batch_torch, elite_returns_torch)
            loss = policy_loss + self._value_loss_coef * value_loss

            # Use the optimizer to train both models.
            self._optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(
                list(self._policy_network.parameters()) + list(self._value_network.parameters()),
                0.5,
            )
            self._optimizer.step()

        # Update the random action probability through the random action mechanism, and then update
        # the best score and best graph.
        self._random_action_mechanism.step(
            previous_best_score=self._best_score, current_best_score=new_best_score
        )
        self._best_score = new_best_score
        self._best_graph = new_best_graph

        # Increment the number of executed iterations of the learning process.
        self._step_count += 1

    @property
    def step_count(self) -> Optional[int]:
        return self._step_count

    @property
    def best_score(self) -> Optional[float]:
        return self._best_score

    @property
    def best_graph(self) -> Optional[Graph]:
        return self._best_graph
